from argparse import Namespace

from pyschism.mesh.hgrid import Hgrid
from pyschism.mesh.vgrid import Vgrid


class VgridCli:

    def __init__(self, args: Namespace):
        hgrid = Hgrid.open(args.hgrid, crs=args.hgrid_crs)
        vgrid = Vgrid.from_binary(hgrid)
        vgrid.write('/tmp/temp_vgrid.in')

    @staticmethod
    def add_subparser_action(subparsers):
        add_vgrid_options_to_parser(subparsers.add_parser('vgrid'))


def add_vgrid_options_to_parser(parser):
    parser.add_argument('hgrid')
    parser.add_argument('--hgrid-crs')
    parser.add_argument(
        "--overwrite", action="store_true",
        help="Allow overwrite of output file.")



from abc import abstractmethod, abstractproperty

from pyschism.forcing.base import ModelForcing


class BoundaryForcing(ModelForcing):

    @abstractmethod
    def get_boundary_string(self, boundary) -> str:
        pass

    @abstractproperty
    def forcing_digit(self):
        """ bctype int value """
        raise NotImplementedError


Bctype = BoundaryForcing



from abc import abstractmethod
import logging
from time import time


from numba import prange, jit
import numpy as np

from pyschism.mesh.gridgr3 import Gr3Field


logger = logging.getLogger(__name__)


class Nudge(Gr3Field):
    def __init__(self, bctides, data_source, rlmax=1.5, rnu_day=0.25):

        @jit(nopython=True, parallel=True)
        def compute_nudge(lon, lat, opbd, out):

            rnu_max = 1.0 / rnu_day / 86400.0

            for idn in prange(lon.shape[0]):
                if idn in opbd:
                    rnu = rnu_max
                    distmin = 0.0
                else:
                    distmin = np.finfo(np.float64).max
                    for j in opbd:
                        rl2 = np.sqrt(
                            np.square(lon[idn] - lon[j]) + np.square(lat[idn] - lat[j])
                        )
                        if rl2 < distmin:
                            distmin = rl2
                rnu = 0.0
                if distmin <= rlmax:
                    rnu = (1 - distmin / rlmax) * rnu_max
                out[idn] = rnu

        out = np.zeros(bctides.hgrid.values.shape)
        elnode = bctides.hgrid.elements.array
        xy = bctides.hgrid.get_xy(crs="epsg:4326")
        opbd = []
        for boundary in bctides.gdf.itertuples():
            forcing = getattr(boundary, self.bctype)
            if forcing.nudge is True:
                opbd.extend(list(boundary.indexes))
        opbd = np.array(opbd)

        if len(opbd) > 0:
            logger.info(f"Begin compute_nudge for {self.name}.")
            start = time()
            compute_nudge(xy[:, 0], xy[:, 1], opbd, out)
            
            logger.info(f"Get the indexes for nudge.")
            idxs_nudge = np.zeros(out.shape, dtype=int)
            idxs = np.where(out > 0)[0]
            idxs_nudge[idxs] = 1
            idxs = np.where(np.max(out[elnode], axis=1) > 0)[0]
            fp = elnode[idxs, -1] < 0
            idxs_nudge[elnode[idxs[fp], :3]] = 1
            idxs_nudge[elnode[idxs[~fp], :]] = 1
            idxs = np.where(idxs_nudge == 1)[0]
            include = idxs
            logger.info(f'The shape of include is {len(include)}')
 
            logger.info(f"compute_nudge took {time()-start} seconds.")
        super().__init__(
            nodes={i: (coord, out[i]) for i, coord in enumerate(bctides.hgrid.coords)},
            elements=bctides.hgrid.elements.elements,
            description=f"{rlmax}, {rnu_day}",
            crs=bctides.hgrid.crs,
        )

        self.data_source = data_source

    #def write()

    @property
    @abstractmethod
    def name(self):
        """ """

    @property
    @abstractmethod
    def bctype(self):
        """ """


class TEM_Nudge(Nudge):
    @property
    def name(self):
        return "temperature"

    @property
    def bctype(self):
        return "itetype"


class SAL_Nudge(Nudge):
    @property
    def name(self):
        return "salinity"

    @property
    def bctype(self):
        return "isatype"


TempNudge = TEM_Nudge
SaltNudge = SAL_Nudge



from argparse import Namespace

import matplotlib.pyplot as plt

from pyschism.mesh import Hgrid


class HgridCli:

    def __init__(self, args: Namespace):

        if args.action == 'plot':
            plot_hgrid(args)

        else:
            raise NotImplementedError(f'Unhandled CLI action: {args.action}.')

    @staticmethod
    def add_subparser_action(subparsers):
        add_hgrid_options_to_parser(subparsers.add_parser('hgrid'))


def add_hgrid_options_to_parser(parser):
    actions = parser.add_subparsers(dest='action')
    add_plot_hgrid(actions)


def plot_hgrid(args):
    hgrid = Hgrid.open(args.path, crs=args.crs)
    ax = None
    if not args.no_topobathy:
        ax = hgrid.make_plot(
                vmin=args.vmin,
                vmax=args.vmax,
            )
    if args.show_elements:
        ax = hgrid.triplot(axes=ax)

    # if args.plot_boundaries:
    #     hgrid.plot_boundaries(axes=ax)

    plt.show()


def add_plot_hgrid(subparser):
    plot = subparser.add_parser('plot')
    plot.add_argument('path')
    plot.add_argument('--crs')
    plot.add_argument('--vmin')
    plot.add_argument('--vmax')
    plot.add_argument('--show-elements', action='store_true')
    plot.add_argument('--no-topobathy')



#! /usr/bin/env python
import unittest
import os


class OutputsApiTestCase(unittest.TestCase):

    def test_outputs_api(self):
        test_data = f'{os.getenv("ICOGS3D_HGRID")}'
        # test_data = "/sciclone/pscr/lcui01/ICOGS3D_dev/outputs_run3c"
        outputs = OutputCollection(test_data)
        # anim = outputs.elev.animation(vmin=-3, vmax=3, fps=6)
        from time import time
        start_total = time()
        print('Begin time aggregation.')
        for dt in outputs.elev.flattened_timevector:
            start_local = time()
            outputs.elev.aggregate(dt)
            print(f'took {time()-start_local} to aggregate elev for time {dt}.')
        print(f'Total aggregation time for variable elev took {time()-start_total}')


if __name__ == '__main__':
    unittest.main()



from argparse import Namespace

from pyschism.cmd.fgrid import manning


class FgridCli:

    def __init__(self, args: Namespace):

        if args.action == 'manning':
            manning.ManningsNCli(args)

        else:
            raise NotImplementedError(f'Unhandled CLI action: {args.action}.')

    @staticmethod
    def add_subparser_action(subparsers):
        add_fgrid_options_to_parser(subparsers.add_parser('fgrid'))


def add_fgrid_options_to_parser(parser):
    actions = parser.add_subparsers(dest='action')
    manning.add_manning(actions)



from pyschism.forcing.bctides.tides import Tides
from pyschism.forcing.bctides.iettype import Iettype
from pyschism.forcing.bctides.ifltype import Ifltype
from pyschism.forcing.bctides.isatype import Isatype
from pyschism.forcing.bctides.itetype import Itetype
from pyschism.forcing.bctides.itrtype import Itrtype
from pyschism.forcing.bctides.bctides import Bctides

__all__ = ["Bctides", "Tides", "Iettype", "Ifltype", "Isatype", "Itetype", "Itrtype"]



from datetime import datetime, timedelta
import pickle
import copy
from numbers import Integral, Real

from pandas import read_csv, to_datetime, to_numeric
import pandas as pd
from matplotlib import pyplot
import numpy

def running_mean1(X, N):
    Y = X * numpy.nan
    if (X.ndim > 1):  # multi-dimensional
        for i, x in enumerate(X.transpose()):
            Y[:, i] = numpy.convolve(x, numpy.ones((N,))/N)[(N-1):]
    else:  # 1-d array
        Y = numpy.convolve(X, numpy.ones((N,))/N)[(N-1):]

    return Y

def running_mean(X, N):
    cumsum = numpy.cumsum(numpy.insert(X, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)


class TimeHistory():


    """Class for manipulating *.th"""

    def __init__(self, file_name=None, start_time_str="2000-01-01 00:00:00", mask_val=None, sec_per_time_unit=1.0, data_array=None, columns=None):

        if mask_val is None:
            mask_val = -9999999
        self.mask_val = mask_val

        """initialize """
        self.source_file = file_name
        self.start_time_str = start_time_str
        self.n_time = -99999999
        self.n_station = -99999999
        self.time = None
        self.data = None
        self.delta_t = -99999999.9
        self.datetime = None
        self.sec_per_time_unit = sec_per_time_unit

        if file_name is None:
            self.df = pd.DataFrame(data_array)
        else:
            self.df = read_csv(file_name, delim_whitespace=True,  index_col=False, header=None)
        if columns is not None:
            self.df.columns = columns

        self.df.rename(columns={0: 'datetime'}, inplace=True)
        self.df_propagate()  # populate vars from dataframe

    def df_propagate(self):
        [self.n_time, self.n_station] = list(self.df.shape)
        self.n_station -= 1  # first col is time
        if isinstance(self.df['datetime'][0], (Real, Integral)):
            self.time = self.df['datetime'].to_numpy()
            self.datetime = [datetime.strptime(self.start_time_str, '%Y-%m-%d %H:%M:%S')
                             + timedelta(seconds=x*self.sec_per_time_unit) for x in self.time]
            self.df['datetime'] = self.datetime
        elif isinstance(self.df['datetime'][0], datetime):
            time_delta = self.df['datetime'] - self.df['datetime'][0]
            self.time = numpy.array([x.total_seconds() for x in time_delta])
            self.datetime = self.df['datetime']
        else:
            raise Exception('unknown datatype for the first column, neither datetime or float')


        # mask invalid values
        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        print("Number of times: " + str(self.n_time))
        print("Number of stations: " + str(self.n_station))

        self.delta_t = self.time[1] - self.time[0]
        if self.delta_t < 50:  # probably in days
            self.time *= 86400.0
            self.delta_t *= 86400.0
            print("Time unit converted from days to seconds")
        else:
            print("Time unit: seconds")

    def get_running_mean(self, station_idx, window):
        """ sort time series by time average"""
        if station_idx is None:
            valid_idx = range(0, self.n_station)
        else:
            valid_idx = station_idx
        print(valid_idx)

        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        data_rm = running_mean(self.data[:, valid_idx], window)

        return [data_rm]

    def get_time_average(self, station_idx, start_time_str=None, end_time_str=None):
        if station_idx == []:
            valid_idx = range(0, self.n_station)
        else:
            valid_idx = station_idx
        print(valid_idx)

        if start_time_str is None:
            idx1 = 0
        else:
            idx1 = self.get_snapshot(start_time_str)

        if end_time_str is None:
            idx2 = self.n_time
        else:
            idx2 = self.get_snapshot(end_time_str)

        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        data_mean = numpy.mean(self.data[idx1:idx2, valid_idx], axis=0)

        return data_mean

    def sort_time_average(self, station_idx):
        """ sort time series by time average"""
        if station_idx == []:
            valid_idx = range(0, self.n_station)
        else:
            valid_idx = station_idx
        print(valid_idx)

        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        data_mean = numpy.mean(self.data[:, valid_idx], axis=0)
        id_sort = numpy.argsort(data_mean)
        print("Preview on sorted: ")
        print(data_mean[id_sort[-1:-11:-1]])

        return [id_sort, data_mean]

    def get_snapshot(self, this_time_str=None, this_time_s_1970=None):
        """ get a snapshot closest to the input time"""
        if this_time_str is None:
            raise Exception("needs this_time_str to get snapshot")
        this_time_sec = timedelta(datetime.strftime(this_time_str, "%Y-%m-%d %H:%M:%S")
                                  - self.datetime[0]).seconds

        t_idx = numpy.argmin(numpy.abs(self.time - this_time_sec))
        return(t_idx)

    def get_max_station(self):
        """ get max among all stations """
        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        data_sum = numpy.sum(self.data, axis=0)
        max_idx = numpy.argmax(data_sum)
        print("Station with the largest time-average value: " + str(max_idx))

        return max_idx

    def simple_plot(self, idx, **args):
        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        pyplot.plot(self.datetime, self.data[:, idx], **args)

    def time_unit_conversion(self, unit_conv):
        """convert time unit"""
        """Col 0 is time: unit_conv from sec to day is 1/86400"""
        self.time = self.time * unit_conv

    def data_unit_conversion(self, unit_conv):
        """convert data unit"""
        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan
        self.data = self.data * unit_conv

    def export_subset(self, station_idx=None, time_idx=None, i_reset_time=False, subset_filename=None):
        import os
        """extract a subset from the original *.th"""
        """by station_idx and time_idx"""
        """[] idx means no operation"""
        """if i_reset_time == True, then the subset *.th starts from 0 time"""

        if subset_filename is None:
            subset_filename = os.path.abspath(self.source_file) + '.subset'

        if station_idx is None:
            station_idx = numpy.array(range(self.n_station))
        else:
            station_idx = numpy.array(station_idx)

        if time_idx is None:
            time_idx = numpy.array(range(self.n_time))
        else:
            time_idx = numpy.array(time_idx)

        self.data = self.df.iloc[:, 1:].to_numpy(dtype=float)
        self.data[abs(self.data-float(self.mask_val)) < 1e-5] = numpy.nan

        sub_data = copy.deepcopy(self.data)
        sub_time = copy.deepcopy(self.time)
        if (len(station_idx) > 0):
            sub_data = sub_data[:, station_idx]
        if (len(time_idx) > 0):
            sub_data = sub_data[time_idx, :]
            sub_time = self.time[time_idx]

        if i_reset_time:
            new_start_time_str = datetime.strftime(datetime.strptime(self.start_time_str, "%Y-%m-%d %H:%M:%S") +timedelta(seconds=sub_time[0]*self.sec_per_time_unit), "%Y-%m-%d %H:%M:%S")
            sub_time = sub_time - sub_time[0]
        else:
            new_start_time_str = self.start_time_str

        with open(subset_filename, 'w') as fout:
            for i, _ in enumerate(sub_time):
                fout.write(str(sub_time[i]) + " " +
                           ' '.join(map(str, sub_data[i, :])) +
                           "\n")
        
        return TimeHistory(file_name=subset_filename, start_time_str=new_start_time_str, mask_val=self.mask_val, sec_per_time_unit=self.sec_per_time_unit,
                           data_array=None, columns=None)

    def writer(self, out_file_name):
        self.df_propagate()
        """ write self """
        with open(out_file_name, 'w') as fout:
            for i, _ in enumerate(self.time):
                fout.write(str(self.time[i]) + " " +
                           ' '.join(map(str, self.data[i, :])) +
                           "\n")

    @classmethod
    def timestamp_microsecond(cls, epoch, utc_time):
        """ convert date to microseconds """
        time_diff = utc_time - epoch
        assert time_diff.resolution == timedelta(microseconds=1)
        return (time_diff.days * 86400 + time_diff.seconds) * 10**6 + time_diff.microseconds



from argparse import Namespace

from pyschism.outputs.stations import StationsOutput


class StationsCli:

    def __init__(self, args: Namespace):

        if args.action == 'plot':
            StationsOutput(
                    args.outputs_directory,
                    args.stations_file,
            ).plot(args.variable, args.station_index, show=True)

        else:
            raise NotImplementedError(f'Unhandled CLI action: {args.action}.')

    @staticmethod
    def add_subparser_action(subparsers):
        add_stations(subparsers.add_parser('stations'))


def add_stations(parser):
    # define subparser action
    actions = parser.add_subparsers(dest='action')
    # plotting action
    plot = actions.add_parser('plot')
    plot.add_argument('variable')
    plot.add_argument('station_index', type=int)
    plot.add_argument('outputs_directory')
    plot.add_argument('--stations-file')



from pyschism.forcing.nws.nws2 import NWS2
from pyschism.forcing.nws.nws2.gfs import GlobalForecastSystem
from pyschism.forcing.nws.nws2.hrrr import HRRR
from pyschism.forcing.nws.nws2.era5 import ERA5
from pyschism.forcing.nws.best_track import BestTrackForcing
from pyschism.forcing.nws.base import NWS


GFS = GlobalForecastSystem


__all__ = [
    'NWS',
    "NWS2",
    'GlobalForecastSystem',
    'GFS',
    'HRRR',
    'ERA5',
    'BestTrackForcing',
]



#! /usr/bin/env python
from datetime import timedelta
import logging
import pathlib
import shutil

from pyschism import dates
from pyschism.driver import ModelConfig
from pyschism.forcing.bctides import iettype, ifltype
from pyschism.forcing.nws import NWS2, GFS, HRRR
from pyschism.forcing.source_sink import NWM
from pyschism.mesh import Hgrid

SCHISM_BINARY = 'pschism_TVD-VL'


def test_barotropic_forecast():

    test_output_directory = pathlib.Path("/tmp/test")
    if test_output_directory.exists():
        shutil.rmtree(test_output_directory)

    config = ModelConfig(
        Hgrid.open(
            "https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll",
            crs="epsg:4326",
        ),
        iettype=iettype.Iettype3(database="tpxo"),
        ifltype=ifltype.Ifltype3(database="tpxo"),
        nws=NWS2(GFS(), HRRR()),
        source_sink=NWM(
            # aggregation_radius=4000.
        ),
    )

    config.forcings.nws.sflux_2.inventory.file_interval = timedelta(hours=6)

    # create reference dates
    nearest_cycle = dates.nearest_cycle()
    spinup_time = timedelta(days=1)
    coldstart = config.coldstart(
        start_date=nearest_cycle - spinup_time,
        end_date=nearest_cycle,
        timestep=300.0,
        dramp=spinup_time,
        dramp_ss=spinup_time,
        drampwind=spinup_time,
        nspool=timedelta(hours=1),
        elev=True,
        dahv=True,
    )

    # optionally run or write the coldstart object
    if shutil.which(SCHISM_BINARY) is not None:
        coldstart.run(
            test_output_directory / "coldstart",
            overwrite=True,
        )

    else:
        coldstart.write(test_output_directory / "coldstart", overwrite=True)

    hotstart = config.hotstart(
        coldstart,
        end_date=timedelta(days=1) + timedelta(hours=23.0),
        timestep=300.0,
        nspool=timedelta(hours=1.0),
        elev=True,
        dahv=True,
    )

    if shutil.which(SCHISM_BINARY) is not None:
        hotstart.run(
            test_output_directory / "hotstart",
            overwrite=True,
        )

    else:
        hotstart.write(test_output_directory / "hotstart", overwrite=True)


if __name__ == "__main__":

    logging.basicConfig(
        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
        force=True,
    )
    logging.captureWarnings(True)
    logging.getLogger("pyschism").setLevel(logging.DEBUG)
    test_barotropic_forecast()



from pyschism.mesh.hgrid import Gr3

if __name__ == '__main__':
    hgrid = Gr3.open('./hgrid.gr3', crs='epsg:4326')
    grd = hgrid.copy()

    gr3_names = ['albedo', 'diffmax', 'diffmin', 'watertype', 'windrot_geo2proj']
    values = [0.1, 1.0, 1e-6, 1.0, 0.0]

    for name, value in zip(gr3_names, values):
       grd.description = name
       grd.nodes.values[:] = value
       grd.write(f'{name}.gr3', overwrite=True)




import argparse
import logging
import warnings

from pyproj import CRS

from pyschism.forcing.bctides.nudge import Nudge
from pyschism.forcing.baroclinic import GOFS, RTOFS
from pyschism.mesh import Hgrid, Vgrid

logger = logging.getLogger(__name__)


class NudgeCli:
    def __init__(self, args: argparse.Namespace):
        data_source = {
            "gofs": GOFS,
            "rtofs": RTOFS,
        }[args.baroclinic_database]()
        nudge = Nudge(args.hgrid, rlmax=args.rlmax, rnu_day=args.rnu_day)
        temp = nudge(data_source.temperature, args.vgrid)
        temp.write('TEM_nudge.gr3')
        salt = nudge(data_source.salinity, args.vgrid)
        salt.write('SAL_nudge.gr3')


def add_nudge(subparsers):
    nudge = subparsers.add_parser("nudge")
    nudge.add_argument(
        "hgrid",
        action=HgridAction,
    )
    nudge.add_argument(
        "vgrid",
        action=VgridAction,
    )
    nudge.add_argument(
        "--hgrid-crs",
        action=HgridCrsAction
    )
    nudge.add_argument('--rlmax', type=float, default=1.5)
    nudge.add_argument('--rnu_day', type=float, default=0.25)
    nudge.add_argument(
        '--hycom',
        '--baroclinic-database',
        choices=['rtofs', 'gofs'],
        dest='baroclinic_database',
        default='gofs',
    )
    nudge.add_argument(
        "--log-level",
        choices=[name.lower() for name in logging._nameToLevel],
        default="warning",
    )


class HgridAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            hgrid = Hgrid.open(values)
        if len(hgrid.boundaries.open) == 0:
            raise TypeError(f"Hgrid provided {values} contains no open boundaries.")
        setattr(namespace, self.dest, hgrid)


class HgridCrsAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values is not None:
            namespace.hgrid.nodes._crs = CRS.from_user_input(values)
        setattr(namespace, self.dest, values)


class VgridAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, Vgrid.open(values))



#! /usr/bin/env python
import argparse
# from datetime import datetime
import logging
import os
import pathlib
# import shutil
import tarfile
import tempfile
import unittest
import urllib.request

from pyschism.outputs.outputs import OutputCollection


# logging.basicConfig(level=logging.INFO, force=True)


class OutputsApiTestCase(unittest.TestCase):

    def test_outputs_api(self):
        test_data = f"{os.getenv('HOME')}/pyschism/forecast/forecasts/2021-04-02T00:00:00+00:00/outputs/"
        # test_data = "/sciclone/pscr/lcui01/ICOGS3D_dev/outputs_run3c"
        outputs = OutputCollection(test_data)
        # anim = outputs.elev.animation(vmin=-3, vmax=3, fps=6)
        from time import time
        start_total = time()
        print('Begin time aggregation.')
        for dt in outputs.elev.flattened_timevector:
            start_local = time()
            outputs.elev.aggregate(dt)
            print(f'took {time()-start_local} to aggregate elev for time {dt}.')
        print(f'Total aggregation time for variable elev took {time()-start_total}')



if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
# import argparse
# from datetime import datetime
# import logging
import os
# import pathlib
# import shutil
# import tarfile
# import tempfile
import unittest
# import urllib.request

from pyschism.outputs.outputs import OutputCollection


# logging.basicConfig(level=logging.INFO, force=True)

class OutputsApiTestCase(unittest.TestCase):

    def test_outputs_api(self):
        test_data = f"{os.getenv('HOME')}/pyschism/forecast/forecasts/2021-04-02T00:00:00+00:00/outputs/"
        # test_data = "/sciclone/pscr/lcui01/ICOGS3D_dev/outputs_run3c"
        print("init outputs")
        from time import time
        start_total = time()
        outputs = OutputCollection(test_data)
        print(f'init took {time()-start_total}')
        outputs.elev.aggregate_parallel(nprocs=16)
        # anim = outputs.elev.animation(vmin=-3, vmax=3, fps=6)
        # for dt in outputs.elev.flattened_timevector:
        #     start_local = time()
        #     outputs.elev.aggregate(dt)
        #     print(f'took {time()-start_local} to aggregate elev for time {dt}.')
        # print(f'Total aggregation time for variable elev took {time()-start_total}')


if __name__ == '__main__':
    unittest.main()



from datetime import datetime, timedelta
import pathlib
import os
from typing import Union

import numpy as np
import pandas as pd
import xarray as xr
import fsspec
import zarr

from pyschism.dates import nearest_cycle

class AWSZarrInventory:
    '''
    12/21/2021, L. Cui
    There is an issue with DSWRF in this dataset. After communicating with 
    MesoWest group, it is confirmed that they are unable to successfully 
    generate zarr-formatted output for this vaiable for the zarr-formatted
    forecast files. At this point, I'll keep this script in the repo. 
    '''

    def __init__(self, bbox=None):
        self.bbox = bbox

    def gen_sflux(self, outdir: Union[str, os.PathLike], start_date=None, air=True, prc=True, rad=True):

        start_date = nearest_cycle() if start_date is None else start_date
        airVars = {'stmp': ['TMP', '2m_above_ground'],
                   'spfh': ['SPFH', '2m_above_ground'],
                   'uwind': ['UGRD', '10m_above_ground'],
                   'vwind': ['VGRD', '10m_above_ground'],
                   'prmsl': ['MSLMA', 'mean_sea_level']}
        prcVars = {'prate': ['PRATE', 'surface']}
        radVars = {'dlwrf': ['DLWRF', 'surface'],
                   'dswrf': ['DSWRF', 'surface']}

        lon, lat, idx_ymin, idx_ymax, idx_xmin, idx_xmax = self.modified_latlon()

        path = pathlib.Path(outdir)
        path.mkdir(parents=True, exist_ok=True)

        #for cycle in range(0, 24, 6):

        cycle = start_date.hour
        base_url = f's3://hrrrzarr/sfc/{start_date.strftime("%Y%m%d")}/' \
            + f'{start_date.strftime("%Y%m%d")}_{cycle:02d}z_fcst.zarr'

        xdat = []
        lon2 = xr.DataArray(
            data=lon, 
            dims=("ny_grid", "nx_grid",), 
            name="lon",
            attrs=dict(
                long_name = "Longitude",
                standard_name = "longitude",
                units = "degrees_east"
            )
        )
        xdat.append(lon2)
        lat2 = xr.DataArray(
            data=lat, 
            dims=("ny_grid", "nx_grid",), 
            name="lat",
            attrs=dict(
                long_name = "Latitude",
                standard_name = "latitude",
                units = "degrees_north"
            )
        )
        xdat.append(lat2)

        if air:
            for key, value in airVars.items():
                url = f'{base_url}/{value[1]}/{value[0]}/{value[1]}/'
                print(url)
                ds = xr.open_zarr(fsspec.get_mapper(url, anon=True))
                ds = ds.rename({'projection_y_coordinate': 'ny_grid'})
                ds = ds.rename({'projection_x_coordinate': 'nx_grid'})
                val = ds[value[0]][:, idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                xdat.append(xr.DataArray(val, dims=("time", "ny_grid", "nx_grid",), name=key))

        if prc:
            for key, value in prcVars.items():
                url = f'{base_url}/{value[1]}/{value[0]}/{value[1]}/'
                print(url)
                ds = xr.open_zarr(fsspec.get_mapper(url, anon=True))
                ds = ds.rename({'projection_y_coordinate': 'ny_grid'})
                ds = ds.rename({'projection_x_coordinate': 'nx_grid'})
                val = ds[value[0]][:, idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                xdat.append(xr.DataArray(val, dims=("time", "ny_grid", "nx_grid",), name=key))

        if rad:
            for key, value in radVars.items():
                url = f'{base_url}/{value[1]}/{value[0]}/{value[1]}/'
                print(url)
                ds = xr.open_zarr(fsspec.get_mapper(url, anon=True))
                ds = ds.rename({'projection_y_coordinate': 'ny_grid'})
                ds = ds.rename({'projection_x_coordinate': 'nx_grid'})
                val = ds[value[0]][:, idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                #DSWRF are all junk values in this dataset
                #if np.sum(np.isnan(val)) > 0:
                #    val[:] = 0.0
                xdat.append(xr.DataArray(val, dims=("time", "ny_grid", "nx_grid",), name=key))
        
        time = ds['time'].values.astype('float32')
        time = (time+1)/24
        bdate = pd.to_datetime(start_date + timedelta(hours=cycle)).strftime('%Y %m %d %H').split(' ')
        bdate = [int(q) for q in bdate[:4]] + [0]
        xdat.append(
            xr.DataArray(
                time, dims=('time'), 
                name='time2',
                attrs=dict(
                    long_name="Time",
                    standard_name = "time",
                    base_date = bdate,
                    units = f"days since {start_date.year}-{start_date.month}-{start_date.day} {cycle:02d}:00 UTC"
                )))
        fout = xr.merge(xdat)
        fout=fout.rename_vars({'time2': 'time'})
        fout.to_netcdf(path / f'hrrr_{start_date.strftime("%Y%m%d")}{cycle:02d}.nc', 'w', unlimited_dims='time')

    def modified_latlon(self):
        latlon_h5_file = "HRRR_latlon.h5"
        ds = xr.open_dataset(latlon_h5_file)
        lon = ds['longitude'].astype("float32")
        lat = ds['latitude'].astype("float32")
        xmin = self.bbox.xmin
        xmax = self.bbox.xmax
        ymin = self.bbox.ymin
        ymax = self.bbox.ymax
        lon_idxs=(lon.values >= xmin) & (lon.values <= xmax)
        lat_idxs=(lat.values >= ymin) & (lat.values <= ymax+2.0)
        idxs = lon_idxs & lat_idxs
        idxs = np.argwhere(idxs)
        idx_ymin = np.min(idxs[:,0])
        idx_ymax = np.max(idxs[:,0])
        idx_xmin = np.min(idxs[:,1])
        idx_xmax = np.max(idxs[:,1])
        #print(f'idx_ymin is {idx_ymin}, idx_ymax is {idx_ymax}, idx_xmin is {idx_xmin}, idx_xmax is {idx_xmax}')
        return lon[idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1], lat[idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1], idx_ymin, idx_ymax, idx_xmin, idx_xmax



from abc import abstractmethod
from enum import Enum
from typing import List

from pyschism.forcing.bctides.bctypes import Bctype
from pyschism.forcing.bctides.tides import Tides
from pyschism.forcing import hycom


class Iettype(Bctype):

    @property
    @abstractmethod
    def iettype(self) -> int:
        pass

    @property
    def forcing_digit(self):
        return self.iettype


class UniformTimeHistoryElevation(Iettype):
    def __init__(self, time_history: List[List[float]], boundaries):
        self.time_history = time_history
        self.boundaries = boundaries

    def get_boundary_string(self, hgrid, boundary):
        return ""

    @property
    def iettype(self) -> int:
        return 1


class ConstantElevation(Iettype):
    def __init__(self, values: List[float]):
        self.values = List[float]

    def get_boundary_string(self, hgrid, boundary):
        return f"{self.values[boundary.index]:G}"

    @property
    def iettype(self) -> int:
        return 2


class TidalElevation(Iettype):
    def __init__(self, constituents='all', database="hamtide"):
        self.tides = Tides(tidal_database=database, constituents=constituents)
        
    def add_constituent(
            self,
            name: str,
            amplitude,
            angular_frequency,
            phase=0.,
            **kwargs,
    ):
        self.tides.add_constituent(
            name=name,
            angular_frequency=angular_frequency,
            elevation_amplitude=amplitude,
            elevation_phase=phase,
            velocity_amplitude=self.tides._amplitudes['velocity'].get(name, 0.),
            velocity_phase=self.tides._phases['velocity'].get(name, 0.),
            **kwargs
        )

    def get_boundary_string(self, hgrid, boundary, global_constituents: List[str] = None):
        f = []
        crs = 'epsg:4326' if hgrid.crs is not None else None
        if global_constituents is None:
            for constituent in self.tides.get_active_forcing_constituents():
                f.append(f"{constituent}")
                vertices = hgrid.get_xy(crs=crs)[boundary.indexes, :]
                amp, phase = self.tides.get_elevation(constituent, vertices)
                for i in range(len(boundary.indexes)):
                    f.append(f"{amp[i]:.8e} {phase[i]:.8e}")
        else:
            required_constituent = self.tides.get_active_forcing_constituents()
            for constituent in global_constituents:
                f.append(f"{constituent}")
                if constituent not in required_constituent:
                    for i in range(len(boundary.indexes)):
                        f.append(f"{0:.8e} {0:.8e}")
                else:
                    vertices = hgrid.get_xy(crs=crs)[boundary.indexes, :]
                    amp, phase = self.tides.get_elevation(constituent, vertices)
                    for i in range(len(boundary.indexes)):
                        f.append(f"{amp[i]:.8e} {phase[i]:.8e}")

        return "\n".join(f)

    @property
    def iettype(self):
        return 3


class HarmonicElevation(TidalElevation):

    def __init__(self):
        self.tides = Tides(constituents=None)

class SpatiallyVaryingTimeHistoryElevation(Iettype):
    class BaroclinicDatabases(Enum):
        #RTOFS = hycom.RTOFS
        GOFS = hycom.GOFS

    def __init__(self, data_source="gofs"):

        if isinstance(data_source, str):
            data_source = self.BaroclinicDatabases[data_source.upper()].value()

        if not isinstance(data_source, hycom.Hycom):
            raise TypeError(
                "Argument data_source must be of type str or type "
                f"{type(hycom.Hycom)}, not type {type(data_source)}."
            )

        self.data_component: hycom.HycomComponent = data_source.elevation

    def get_boundary_string(self, hgrid, boundary, global_constituents: List[str] = None):
        return ""

    def write(self, path, hgrid, start_date, run_days, overwrite: bool = False):
        self.data_component.write(path, hgrid, start_date, run_days, overwrite)

    @property
    def iettype(self):
        return 4


class TidalAndSpatiallyVaryingElevationTimeHistory(Iettype):
    def __init__(
        self, iettype3: TidalElevation, iettype4: SpatiallyVaryingTimeHistoryElevation
    ):
        self.iettype3 = iettype3
        self.iettype4 = iettype4

    def get_boundary_string(self, hgrid, boundary, global_constituents: List[str] = None):
        return self.iettype3.get_boundary_string(hgrid, boundary, global_constituents)

    def write(self, path, hgrid, start_date, run_days, overwrite: bool = False):
        self.data_component.write(path, hgrid, start_date, run_days, overwrite)

    @property
    def iettype(self):
        return 5

    @property
    def tides(self):
        return self.iettype3.tides

    @property
    def data_component(self):
        return self.iettype4.data_component


class ZeroElevation(Iettype):
    def __init__(self):
        raise NotImplementedError(f"{self.__class__.__name__}")

    def get_boundary_string(self):
        return ""

    @property
    def iettype(self):
        return -1


class Iettype1(UniformTimeHistoryElevation):
    pass


class Iettype2(ConstantElevation):
    pass


class Iettype3(TidalElevation):
    pass


class Iettype4(SpatiallyVaryingTimeHistoryElevation):
    pass


class Iettype5(TidalAndSpatiallyVaryingElevationTimeHistory):
    pass


class Iettype_1(ZeroElevation):
    pass



from abc import ABC, abstractmethod
from enum import Enum
from functools import lru_cache
import pathlib
import subprocess
import tempfile

import numpy as np

from pyschism.mesh.hgrid import Hgrid


def C_of_sigma(sigma, theta_b, theta_f):
    assert theta_b <= 0. and theta_b <= 1.
    assert theta_f <= 0. and theta_f <= 1.
    A = (1-theta_b)(np.sinh(sigma*theta_f)/np.sinh(theta_f))
    B_1 = np.tanh(theta_f*(sigma+0.5)) - np.tanh(theta_f/2.)
    B = theta_b * (B_1 / (2.*np.tanh(theta_f/2.)))
    return A + B


def eta_of_sigma(sigma):
    return 1 + sigma


def S_to_Z(sigma):
    # eq 3.1
    pass


class VgridType(Enum):

    LSC2 = 1
    SZ = 2

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f'ivcor={name} is not a valid vgrid type.')


class Vgrid(ABC):

    @abstractmethod
    def __str__(self):
        raise NotImplementedError
    
    @staticmethod
    def default():
        return SZ.default() 

    @staticmethod
    def v2d(h_s, ztot, h_c, theta_b, theta_f, sigma):
        return SZ._v2d(h_s, ztot, h_c, theta_b, theta_f, sigma)

    @classmethod
    def from_binary(cls, hgrid, binary='gen_vqs'):
        _tmpdir = tempfile.TemporaryDirectory()
        tmpdir = pathlib.Path(_tmpdir.name)
        hgrid = Hgrid.open(hgrid, crs='EPSG:4326')
        hgrid.write(tmpdir / 'hgrid.gr3')
        subprocess.check_call([binary], cwd=tmpdir)
        return cls.open(tmpdir / 'vgrid.in')

    @staticmethod
    def open(path):
        '''
        Based on:
        https://github.com/wzhengui/pylibs/blob/master/Utility/schism_file.py
        '''
        with open(path) as f:
            return VgridTypeDispatch[VgridType(
                int(f.read().strip().split()[0])).name].value.open(path)

    @abstractmethod
    def get_xyz(self, gr3, crs=None):
        pass

    def write(self, path, overwrite=False):
        path = pathlib.Path(path)
        if path.is_file() and not overwrite:
            raise Exception(
                'File exists, pass overwrite=True to allow overwrite.')

        with open(path, 'w') as f:
            f.write(str(self))

    @property
    def ivcor(self):
        return VgridType[self.__class__.__name__].value

    @property
    @abstractmethod
    def nvrt(self):
        raise NotImplementedError

    @lru_cache(maxsize=1)
    def is2D(self):
        if isinstance(self, SZ):
            if str(self) == str(SZ.default()):
                return True
        return False

    def is3D(self):
        return ~self.is2D()


class LSC2(Vgrid):

    def __init__(self, sigma):
        self.sigma = sigma

    def __str__(self):
        f = [
            f'{self.ivcor}',
            f'{self.nvrt}',
        ]
        for i, row in enumerate(self.sigma):
            kbp = int((row == -1).sum())
            line = [
                f'{i+1}'.rjust(11),
                f'{kbp}'.rjust(11),
                7*' ',
                '-1.000000',
            ]
            for value in row:
                if value != -1:
                    line.append(7*' ')
                    line.append(f'{value:6f}')

            f.append(' '.join(line))
        return '\n'.join(f)

    def get_xyz(self, gr3, crs=None):
        xy = gr3.get_xy(crs)
        z = gr3.values[:, None]*self.sigma
        x = np.tile(xy[:, 0], (z.shape[1],))
        y = np.tile(xy[:, 0], (z.shape[1],))
        return np.vstack([x, y, z.flatten()]).T

    @classmethod
    def open(cls, path):

        path = pathlib.Path(path)

        with open(path) as f:
            lines = f.readlines()

        ivcor = int(lines[0].strip().split()[0])
        if ivcor != 1:
            raise TypeError(f'File {path} is not an LSC2 grid (ivcor != 1).')

        nvrt = int(lines[1].strip().split()[0])

        sline = np.array(lines[2].split()).astype('float')        
        if sline.min() < 0:
            #old version
            kbp = np.array([int(i.split()[1])-1 for i in lines[2:]])
            sigma = -np.ones((len(kbp), nvrt))

            for i, line in enumerate(lines[2:]):
                sigma[i, kbp[i]:] = np.array(
                    line.strip().split()[2:]).astype('float')

        else:
            #new version
            sline = sline.astype('int')
            kbp = sline-1
            sigma = np.array([line.split()[1:] for line in lines[3:]]).T.astype('float')
            #replace -9. with -1.
            fpm = sigma<-1
            sigma[fpm] = -1

        return cls(sigma)

    @property
    def nvrt(self):
        return self.sigma.shape[1]


class SZ(Vgrid):

    def __init__(self, h_s, ztot, h_c, theta_b, theta_f, sigma):
        self.h_s = h_s
        self.ztot = np.array(ztot)
        self.h_c = h_c
        self.theta_b = theta_b
        self.theta_f = theta_f
        self.sigma = np.array(sigma)

    def __str__(self):
        f = [
            f'{self.ivcor:d} !ivcor',
            f'{self.nvrt:d} {self.kz:d} {self.h_s:G} '
            '!nvrt, kz (# of Z-levels); h_s '
            ' (transition depth between S and Z)',
            'Z levels',
        ]
        for i, row in enumerate(self.ztot):
            f.append(f'{i+1:d} {row:G}')

        f.extend([
            'S levels',
            f'{self.h_c:G} {self.theta_b:G} {self.theta_f:G} '
            ' !h_c, theta_b, theta_f',
            ])
        for i, row in enumerate(self.sigma):
            f.append(f'{i+1:d} {row:G}')
        return '\n'.join(f)

    def get_xyz(self, gr3, crs=None):
        raise NotImplementedError('SZ.get_xyz')

    @classmethod
    def open(cls, path):

        path = pathlib.Path(path)

        with open(path) as f:
            lines = f.readlines()

        ivcor = int(lines[0].strip().split()[0])
        if ivcor != 2:
            raise TypeError(f'File {path} is not an SZ grid (ivcor != 2).')

        nvrt = int(lines[1].strip().split()[0])

        kz, h_s = lines[1].strip().split()[1:3]
        kz = int(kz)
        h_s = float(h_s)

        # read z grid
        ztot = []
        irec = 2
        for i in np.arange(kz):
            irec = irec+1
            ztot.append(lines[irec].strip().split()[1])
        ztot = np.array(ztot).astype('float')
        # read s grid
        sigma = []
        irec = irec+2
        nsigma = nvrt - kz+1
        h_c, theta_b, theta_f = np.array(
            lines[irec].strip().split()[:3]).astype('float')
        for i in np.arange(nsigma):
            irec = irec + 1
            sigma.append(lines[irec].strip().split()[1])
        sigma = np.array(sigma).astype('float')
        return cls(h_s, ztot, h_c, theta_b, theta_f, sigma)

    @classmethod
    def default(cls):
        # h_s, ztot, h_c, theta_b, theta_f, sigma
        return cls(1.e6, [-1.e6], 40., 1., 1.e-4, [-1, 0.])
        #return cls(h_s, ztot, h_c, theta_b, theta_f, sigma)

    @classmethod
    def _v2d(cls, h_s, ztot, h_c, theta_b, theta_f, sigma):
        return cls(h_s, ztot, h_c, theta_b, theta_f, sigma)

    @property
    def kz(self):
        return self.ztot.shape[0]

    @property
    def nvrt(self):
        return self.sigma.shape[0]


class VgridTypeDispatch(Enum):

    LSC2 = LSC2
    SZ = SZ



import os
import pathlib
import subprocess
import tempfile
from typing import Union

import geopandas as gpd
import numpy as np
from shapely.geometry import Polygon, MultiPolygon, Point

from pyschism.forcing.hycom import Hycom
from pyschism.mesh.base import Gr3


class Gr3Field(Gr3):
    @classmethod
    def constant(cls, hgrid, value):
        obj = cls(
            **{
                k: v
                for k, v in hgrid.to_dict().items()
                if k in ["nodes", "elements", "description", "crs"]
            }
        )
        obj.values[:] = value
        obj.description = f"{cls.__name__.lower()} {obj.crs}"
        return obj

    @classmethod
    def default(cls, hgrid):
        raise NotImplementedError(f"No default defined for {cls.__name__}.")

    def add_region(self, region: Union[Polygon, MultiPolygon], value):
        if isinstance(region, Polygon):
            region = [region]
        gdf1 = gpd.GeoDataFrame({"geometry": region}, crs=self.crs)

        points = [Point(*coord) for coord in self.coords]
        gdf2 = gpd.GeoDataFrame(
            {"geometry": points, "index": list(range(len(points)))}, crs=self.crs
        )
        gdf_in = gpd.sjoin(gdf2, gdf1, op="within")
        picks = [i.index for i in gdf_in.itertuples()]
        self.values[picks] = value

    def modify_by_region(self, hgrid, fname, value, depth1, flag):
        '''
        reset (flag==0) or add (flag==1) value to a region
        '''
        lines=[line.strip().split() for line in open(fname, 'r').readlines()]
        data=np.squeeze(np.array([lines[3:]])).astype('float')
        x=data[:,0]
        y=data[:,1]
        coords = list( zip(x, y))
        poly = Polygon(coords)

        #region is in cpp projection 
        gdf1 = gpd.GeoDataFrame(
                {'geometry': [poly]})

        points = [Point(*coord) for coord in self.coords]
        gdf2 = gpd.GeoDataFrame(
                 {'geometry': points, 'index': list(range(len(points)))})
        gdf_in = gpd.sjoin(gdf2, gdf1, op="within")
        picks = [i.index for i in gdf_in.itertuples()]
        if flag == 0:
            self.values[picks] = value
        else:
            picks2 = np.where(-hgrid.values > depth1)
            picks3 = np.intersect1d(picks, picks2)
            self.values[picks3] = self.values[picks3] + value


class Albedo(Gr3Field):
    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, 0.15)


class Diffmax(Gr3Field):
    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, 1.0)


class Diffmin(Gr3Field):
    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, 1e-6)


class Watertype(Gr3Field):
    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, 1.0)


def transform_to_cpp(coords, lonc, latc):
    longitude = list(coords[:, 0] / 180 * np.pi)
    latitude = list(coords[:, 1] / 180 * np.pi)
    radius = 6378206.4
    loncc = lonc / 180 * np.pi
    latcc = latc / 180 * np.pi
    x = [
        radius * (longitude[i] - loncc) * np.cos(latcc)
        for i in np.arange(len(longitude))
    ]
    y = [radius * latitude[i] for i in np.arange(len(latitude))]
    return np.vstack([x, y]).T


class Shapiro(Gr3Field):

    @classmethod
    def slope_filter(cls, hgrid, shapiro_vals1, depths, shapiro_max, threshold_slope, 
        regions=None, shapiro_vals2=None, flags=None, lonc=None, latc=None):
        """
        https://github.com/wzhengui/pylibs/blob/1ee35efaa2d52fa682113126d84846ba33318f99/Utility/schism_file.py#L265
        """
        hgrid = hgrid.copy()
        hgrid.nodes.transform_to_cpp(lonc, latc)
        xy = hgrid.nodes.coords
        x = xy[:, 0]
        y = xy[:, 1]
        dp = -hgrid.nodes.values
        elnode = hgrid.elements.array

        #breakpoint()
        #fp = np.any(elnode.mask, axis=1)
        fp = np.any(elnode.mask)
        if fp:
            fpn = ~fp
        x1 = x[elnode[:, 0]]
        y1 = y[elnode[:, 0]]
        v1 = dp[elnode[:, 0]]
        x2 = x[elnode[:, 1]]
        y2 = y[elnode[:, 1]]
        v2 = dp[elnode[:, 1]]
        x3 = x[elnode[:, 2]]
        y3 = y[elnode[:, 2]]
        v3 = dp[elnode[:, 2]]
        a1 = ((x2-x1)*(y3-y1)-(x3-x1)*(y2-y1))/2
        if fp:
            mask = np.any(elnode.mask, axis=1)
            x4 = x[elnode[:, 3]]
            y4 = y[elnode[:, 3]]
            v4 = dp[elnode[:, 3]]
            x4[mask] = x1[mask]
            y4[mask] = y1[mask]
            v4[mask] = v1[mask]
            a2 = ((x3-x1)*(y4-y1)-(x4-x1)*(y3-y1))/2

        # compute gradients
        dpedx = (v1*(y2-y3)+v2*(y3-y1)+v3*(y1-y2))/(2*a1)
        dpedy = ((x3-x2)*v1+(x1-x3)*v2+(x2-x1)*v3)/(2*a1)
        dpedxy = np.sqrt(dpedx**2+dpedy**2)

        # modify quads
        if fp:
            mask = np.any(elnode.mask, axis=1)
            fpn = ~mask
            dpedx2 = (v1[fpn]*(y3[fpn]-y4[fpn])+v3[fpn]*(y4[fpn]-y1[fpn])+v4[fpn]*(y1[fpn]-y3[fpn]))/(2*a2[fpn])
            dpedy2 = ((x4[fpn]-x3[fpn])*v1[fpn]+(x1[fpn]-x4[fpn])*v3[fpn]+(x3[fpn]-x1[fpn])*v4[fpn])/(2*a2[fpn])
            dpedxy2 = np.sqrt(dpedx2**2+dpedy2**2)
            
            dpedx[fpn] = (dpedx[fpn]+dpedx2)/2
            dpedy[fpn] = (dpedy[fpn]+dpedy2)/2
            dpedxy[fpn] = (dpedxy[fpn]+dpedxy2)/2

        # get node ball information
        nne, ine = hgrid.elements.get_node_ball()
         
        # interpolate into nodes
        dpdxy = []
        for inode in np.arange(len(dp)):
            ind = ine[inode]
            dpdxy.append(np.max(dpedxy[ind]))
        slope = np.array(dpdxy)

        shapiro = shapiro_max*np.tanh(2*slope/threshold_slope)

        # further tweaks on shallow waters
        if len(depths) != len(shapiro_vals1):
            raise Exception(f'lengths of depths {len(depths)} and shapiro_vals1 {len(shapiro_vals1)} inconsistent')
        fp = dp < depths[-1]
        shapiro[fp] = np.maximum(shapiro[fp], np.interp(dp[fp], depths, shapiro_vals1))

        shapiro = cls(
            nodes={id: (xy[i, :], shapiro[i]) for i, id in enumerate(hgrid.nodes.id)},
            elements=hgrid.elements.to_dict(),
            crs=None,
            description=f"shapiro threshold_slope={threshold_slope} lonc={lonc} latc={latc}",
        )

        if regions is not None:
            for reg, value, flag in zip(regions, shapiro_vals2, flags):
                shapiro.modify_by_region(hgrid, reg, value, depths[0], flag)

        return shapiro


class Windrot(Gr3Field):
    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, 0.0)


class IcField(Gr3Field):
    pass


class ElevIc(IcField):
    @classmethod
    def default(cls, hgrid, offset=0.1):
        obj = cls.constant(hgrid, 0.0)
        obj.values[:] = np.maximum(0.0, hgrid.values - offset)
        return obj

class TempIc(IcField):
    @classmethod
    def from_hycom(cls, gr3: Gr3, hycom: Hycom, date):
        obj = cls.constant(gr3, np.nan)
        obj.values[:] = hycom.temperature.interpolate(obj, date)
        return obj


class SaltIc(IcField):
    @classmethod
    def from_hycom(cls, gr3: Gr3, hycom: Hycom, date):
        obj = cls.constant(gr3, np.nan)
        obj.values[:] = hycom.salinity.interpolate(obj, date)
        return obj


class Estuary(Gr3Field):
    pass



import tempfile
from datetime import timedelta, datetime
from pathlib import Path

from stormevents.nhc import VortexTrack

from pyschism.mesh import Hgrid
from pyschism.driver import ModelConfig
from pyschism.forcing.nws import BestTrackForcing


def test_paramwind_from_stormname():

    hgrid = Hgrid.open(
        'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll'
    )

    meteo = BestTrackForcing(storm='Florence2018')

    config = ModelConfig(
        hgrid=hgrid,
        vgrid=None,
        fgrid=None,
        iettype=None,
        ifltype=None,
        nws=meteo,
    )

    driver = config.coldstart(
        start_date=datetime(2018, 9, 8),
        end_date=datetime(2018, 9, 18),
        timestep=timedelta(seconds=150),
        nspool=24,
    )

    with tempfile.TemporaryDirectory() as dn:
        tmpdir = Path(dn)
        driver.write(tmpdir / 'paramwind', overwrite=True)

def test_paramwind_from_file():
    hgrid = Hgrid.open(
        'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll'
    )

    track = VortexTrack.from_storm_name('Florence', 2018)


    with tempfile.TemporaryDirectory() as dn:
        tmpdir = Path(dn)
        track.to_file(tmpdir / 'track.dat')

        meteo = BestTrackForcing.from_nhc_bdeck(nhc_bdeck=tmpdir / 'track.dat')
        config = ModelConfig(
            hgrid=hgrid,
            vgrid=None,
            fgrid=None,
            iettype=None,
            ifltype=None,
            nws=meteo,
        )

        driver = config.coldstart(
            start_date=datetime(2018, 9, 8),
            end_date=datetime(2018, 9, 18),
            timestep=timedelta(seconds=150),
            nspool=24,
        )

        driver.write(tmpdir / 'paramwind', overwrite=True)





from abc import abstractmethod
from enum import Enum

from pyschism.forcing.bctides.bctypes import Bctype
from pyschism.forcing.bctides.tides import Tides
from pyschism.forcing import hycom


class Ifltype(Bctype):

    @property
    @abstractmethod
    def ifltype(self) -> int:
        """Returns integer representig SCHISM ifltype code for bctides.in"""

    @property
    def forcing_digit(self):
        return self.ifltype


class UniformTimeHistoryVelocity(Ifltype):
    def __init__(self, time_history):
        raise NotImplementedError(f"{self.__class__.__name__}")
        self.time_history = time_history

    @property
    def ifltype(self):
        return 1


class ConstantVelocity(Ifltype):
    def __init__(self, value):
        self.value = value

    @property
    def ifltype(self) -> int:
        return 2

    def get_boundary_string(self, *args, **kwargs):
        return f'{self.value}'


class TidalVelocity(Ifltype):
    def __init__(self, constituents="all", database="hamtide"):
        self.tides = Tides(tidal_database=database, constituents=constituents)

    def get_boundary_string(self, hgrid, boundary, global_constituents=None):
        f = []
        if global_constituents is None:
            for constituent in self.tides.get_active_forcing_constituents():
                f.append(f"{constituent}")
                vertices = hgrid.get_xy(crs="EPSG:4326")[boundary.indexes, :]
                uamp, uphase, vamp, vphase = self.tides.get_velocity(constituent, vertices)
                for i in range(len(vertices)):
                    f.append(
                        f"{uamp[i]:.8e} {uphase[i]:.8e} {vamp[i]:.8e} {vphase[i]:.8e}"
                    )
        else:
            required_constituent = self.tides.get_active_forcing_constituents()
            for constituent in global_constituents:
                f.append(f"{constituent}")
                if constituent not in required_constituent:
                    for i in range(len(boundary.indexes)):
                        f.append(f"{0:.8e} {0:.8e} {0:.8e} {0:.8e}")
                else:
                    vertices = hgrid.get_xy(crs="EPSG:4326")[boundary.indexes, :]
                    uamp, uphase, vamp, vphase = self.tides.get_velocity(constituent, vertices)
                    for i in range(len(boundary.indexes)):
                        f.append(f"{uamp[i]:.8e} {uphase[i]:.8e} {vamp[i]:.8e} {vphase[i]:.8e}")

        return "\n".join(f)

    @property
    def ifltype(self):
        return 3


class SpatiallyVaryingTimeHistoryVelocity(Ifltype):
    class BaroclinicDatabases(Enum):
        #RTOFS = hycom.RTOFS
        GOFS = hycom.GOFS

    def __init__(self, data_source="gofs"):
        if isinstance(data_source, str):
            data_source = self.BaroclinicDatabases[data_source.upper()].value()

        if not isinstance(data_source, hycom.Hycom):
            raise TypeError(
                "Argument data_source must be of type str or type "
                f"{type(hycom.Hycom)}, not type {type(data_source)}."
            )

        self.data_component: hycom.HycomComponent = data_source.velocity

    def write(self, path, hgrid, vgrid, start_date, run_days, overwrite: bool = False):
        self.data_component.write(path, hgrid, vgrid, start_date, run_days, overwrite)

    def get_boundary_string(self, *args, **kwargs):
        return ""

    @property
    def ifltype(self):
        return 4


class TidalAndSpatiallyVaryingVelocityTimeHistory(Ifltype):
    def __init__(
        self,
        ifltype3: TidalVelocity,
        ifltype4: SpatiallyVaryingTimeHistoryVelocity,
    ):
        self.ifltype3 = ifltype3
        self.ifltype4 = ifltype4

    def write(self, path, hgrid, vgrid, start_date, run_days, overwrite: bool = False):
        self.data_component.write(path, hgrid, vgrid, start_date, run_days, overwrite)

    @property
    def ifltype(self):
        return 5

    def get_boundary_string(self, *args, **kwargs):
        return self.ifltype3.get_boundary_string(*args, **kwargs)

    @property
    def tides(self):
        return self.ifltype3.tides

    @property
    def data_component(self):
        return self.ifltype4.data_component


class ZeroVelocity(Ifltype):
    def __init__(self):
        raise NotImplementedError(f"{self.__class__.__name__}")

    @property
    def ifltype(self):
        return -1


class InflowVelocity(Ifltype):
    def __init__(self):
        raise NotImplementedError(f"{self.__class__.__name__}")

    @property
    def ifltype(self):
        return -4


class OutflowVelocity(Ifltype):
    def __init__(self):
        raise NotImplementedError(f"{self.__class__.__name__}")

    @property
    def ifltype(self):
        return -5


class Ifltype1(UniformTimeHistoryVelocity):
    pass


class Ifltype2(ConstantVelocity):
    pass


class Ifltype3(TidalVelocity):
    pass


class Ifltype4(SpatiallyVaryingTimeHistoryVelocity):
    pass


class Ifltype5(TidalAndSpatiallyVaryingVelocityTimeHistory):
    pass


class Ifltype_1(ZeroVelocity):
    pass


class Flather(ZeroVelocity):
    pass



import pathlib
import tempfile

from matplotlib.cm import ScalarMappable
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import requests

from pyschism.figures import figure, get_topobathy_kwargs
from pyschism.mesh.parsers import grd
from pyschism.mesh.base import Gr3  # , sort_edges, signed_polygon_area
from pyschism.mesh.boundaries import Boundaries


class Hgrid(Gr3):
    """
    Class that represents the unstructured planar mesh used by SCHISM.
    """

    def __init__(self, *args, boundaries=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._boundaries = Boundaries(self, boundaries)

    @staticmethod
    def open(path, crs=None):
        if str(path).endswith(".ll") and crs is None:
            crs = "epsg:4326"

        try:
            response = requests.get(path)
            response.raise_for_status()
            tmpfile = tempfile.NamedTemporaryFile()
            with open(tmpfile.name, "w") as fh:
                fh.write(response.text)
            _grd = grd.read(pathlib.Path(tmpfile.name), crs=crs)
        except Exception:
            _grd = grd.read(path, crs=crs)

        _grd["nodes"] = {
            id: (coords, -val) for id, (coords, val) in _grd["nodes"].items()
        }

        return Hgrid(**_grd)

    def to_dict(self, boundaries=True):
        _grd = super().to_dict()
        if boundaries is True:
            _grd.update(
                {
                    "nodes": {
                        id: (coord, -val)
                        for id, (coord, val) in self.nodes.to_dict().items()
                    },
                    "boundaries": self.boundaries.data,
                }
            )
            return _grd
        return _grd

    def copy(self):
        return self.__class__(**super().to_dict())

    @figure
    def make_plot(
        self,
        axes=None,
        vmin=None,
        vmax=None,
        show=False,
        title=None,
        # figsize=rcParams["figure.figsize"],
        extent=None,
        cbar_label=None,
        **kwargs
    ):
        if vmin is None:
            vmin = np.min(self.values)
        if vmax is None:
            vmax = np.max(self.values)
        kwargs.update(**get_topobathy_kwargs(self.values, vmin, vmax))
        kwargs.pop("col_val")
        levels = kwargs.pop("levels")
        if vmin != vmax:
            self.tricontourf(axes=axes, levels=levels, vmin=vmin, vmax=vmax, **kwargs)
        else:
            self.tripcolor(axes=axes, **kwargs)
        self.quadface(axes=axes, **kwargs)
        axes.axis("scaled")
        if extent is not None:
            axes.axis(extent)
        if title is not None:
            axes.set_title(title)
        mappable = ScalarMappable(cmap=kwargs["cmap"])
        mappable.set_array([])
        mappable.set_clim(vmin, vmax)
        divider = make_axes_locatable(axes)
        cax = divider.append_axes("bottom", size="2%", pad=0.5)
        cbar = plt.colorbar(mappable, cax=cax, orientation="horizontal")
        cbar.set_ticks([vmin, vmax])
        cbar.set_ticklabels([np.around(vmin, 2), np.around(vmax, 2)])
        if cbar_label is not None:
            cbar.set_label(cbar_label)
        return axes

    @property
    def boundaries(self):
        return self._boundaries

    @property
    def ocean_boundaries(self):
        return self.boundaries.ocean

    @property
    def land_boundaries(self):
        return self.boundaries.land

    @property
    def interior_boundaries(self):
        return self.boundaries.interior



import os
import pathlib
from typing import Union

import f90nml  # type: ignore[import]

from pyschism.enums import NWSType
from pyschism.forcing.nws.base import NWS
from pyschism.forcing.nws.nws2.sflux import SfluxDataset
from pyschism.mesh import gridgr3


SFLUX_DEFAULTS = f90nml.read(pathlib.Path(__file__).parent / "sflux_inputs.txt")


class NWS2(NWS):
    def __init__(
        self,
        sflux_1: SfluxDataset = None,
        sflux_2: SfluxDataset = None,
        windrot: gridgr3.Windrot = None,
    ):
        """Loads SfluxDataset to use as NWS2 input."""

        self.sflux_1 = sflux_1
        self.sflux_2 = sflux_2
        self.windrot = windrot

    def __str__(self):
        data = []
        data = "\n".join(data)
        return f"&sflux_inputs\n{data}/\n"

    @classmethod
    def read(cls, path, sflux_1_glob="*_1.*", sflux_2_glob="*_2.*"):
        path = pathlib.Path(path)
        sflux_2 = list(path.glob(sflux_2_glob))
        return cls(
            sflux_1=SfluxDataset(list(path.glob(sflux_1_glob))),
            sflux_2=SfluxDataset(sflux_2) if len(sflux_2) > 0 else None,
        )

    def write(
        self,
        path: Union[str, os.PathLike],
        start_date=None,
        end_date=None,
        bbox=None,
        overwrite: bool = False,
        windrot: bool = True,
        air=True,
        rad=True,
        prc=True,
    ):

        # write sflux namelist
        path = pathlib.Path(path)
        if path.name != "sflux":
            path /= "sflux"
        path.mkdir(parents=True, exist_ok=overwrite)
        with open(path / "sflux_inputs.txt", "w") as f:
            f.write(str(self))
        # write sflux data
        if self.sflux_1 is not None:
            self.sflux_1.write(
                path,
                1,
                start_date=start_date,
                rnday=end_date,
                air=air,
                rad=rad,
                prc=prc,
                bbox=bbox,
                overwrite=overwrite,
            )
        if self.sflux_2 is not None:
            self.sflux_2.write(
                path,
                2,
                start_date=start_date,
                rnday=end_date,
                air=air,
                rad=rad,
                prc=prc,
                bbox=bbox,
                overwrite=overwrite,
            )
        # # write windrot data
        if windrot is not False and self.windrot is not None:
            windrot = "windrot_geo2proj.gr3" if windrot is True else windrot
            self.windrot.write(path.parent / "windrot_geo2proj.gr3", overwrite)

    @property
    def dtype(self) -> NWSType:
        """Returns the datatype of the object"""
        return NWSType(2)

    @property
    def sflux_1(self) -> SfluxDataset:
        return self._sflux_1

    @sflux_1.setter
    def sflux_1(self, sflux_1):
        if not isinstance(sflux_1, SfluxDataset):
            raise TypeError(
                f"Argument sflux_1 must be of type {SfluxDataset},"
                f" not type {type(sflux_1)}"
            )
        self._sflux_1 = sflux_1

    @property
    def sflux_2(self) -> SfluxDataset:
        return self._sflux_2

    @sflux_2.setter
    def sflux_2(self, sflux_2):
        if sflux_2 is not None:
            if not isinstance(sflux_2, SfluxDataset):
                raise TypeError(
                    f"Argument sflux_2 must be of type {SfluxDataset}, not "
                    f"type {type(sflux_2)}."
                )
        self._sflux_2 = sflux_2

    @property
    def windrot(self):
        return self._windrot

    @windrot.setter
    def windrot(self, windrot: Union[gridgr3.Windrot, None]):
        if not isinstance(windrot, (gridgr3.Windrot, type(None))):
            raise TypeError(
                f"Argument windrot must be of type {gridgr3.Windrot} or None, "
                f"not type {type(windrot)}."
            )
        self._windrot = windrot



from datetime import timedelta
# from enum import Enum
import pathlib
import os
from typing import Union

import f90nml

from pyschism.enums import Stratification


# class IbcType(Enum):
#     BAROCLINIC = 0
#     BAROTROPIC = 1

#     @classmethod
#     def _missing_(self, name):
#         raise ValueError(f'{name} is not a valid integer for ibc. '
#                          'Valid integers are 0 or 1.')


class CORE:
    """Provides error checking implementation for CORE group"""

    mandatory = ["ipre", "ibc", "ibtp", "rnday", "dt", "nspool", "ihfskip"]

    def __init__(
        self,
        ipre: int = 0,
        ibc: int = 0,
        ibtp: int = 0,
        rnday: Union[float, timedelta] = 0.0,
        dt: Union[float, timedelta] = 150.0,
        nspool: Union[int, float, timedelta] = None,
        ihfskip: Union[int, timedelta] = None,
        template: Union[str, os.PathLike, bool] = None,
    ):
        self.ipre = ipre
        self.ibc = ibc
        self.ibtp = ibtp
        self.rnday = rnday
        self.dt = dt
        self.nspool = nspool
        self.ihfskip = ihfskip
        self.template = template

    def __str__(self):
        data = []
        for key, default in self.defaults.items():
            if hasattr(self, key):
                current = getattr(self, key)
            else:
                continue
            if key in self.mandatory:
                data.append(f"  {key}={str(current)}")
            elif default != current:
                data.append(f"  {key}={str(current)}")
        data = "\n".join(data)
        return f"&CORE\n{data}\n/"

    def to_dict(self):
        output = {}
        for key, default in self.template.items():
            if hasattr(self, f"{key}"):
                current = getattr(self, f"{key}")
            else:
                current = None
            if key in self.mandatory:
                output[key] = current
            elif default != current:
                output[key] = current
        return output

    @property
    def ipre(self) -> int:
        return self._ipre

    @ipre.setter
    def ipre(self, ipre: int):
        if ipre not in [0, 1]:
            raise ValueError("Argument to ipre attribute must be 0 or 1")
        self._ipre = ipre

    @property
    def ibc(self):
        return self._ibc

    @ibc.setter
    def ibc(self, ibc: Union[Stratification, int, str]):

        if isinstance(ibc, str):
            ibc = Stratification[ibc.upper()].value

        if isinstance(ibc, Stratification):
            ibc = ibc.value

        if isinstance(ibc, int):
            if ibc not in [0, 1]:
                raise ValueError(
                    "Argument to attribute ibc must be of type "
                    f"{Stratification} or an 0, 1 integer or a string "
                    "'barotropic', 'baroclinic', not type "
                    f"{type(ibc)}."
                )

        self._ibc = ibc

    @property
    def ibtp(self):
        return self._ibtp

    @ibtp.setter
    def ibtp(self, ibtp: int):

        if ibtp not in [0, 1]:
            raise TypeError(
                "Argument to attribute ibtp must be 0 or 1, not " f"{ibtp}."
            )
        if ibtp == 1 and self.ibc == 0:
            raise ValueError("ibtp cannot be set to 1 because ibc is equal to " "zero.")
        self._ibtp = ibtp

    @property
    def rnday(self) -> Union[float, None]:
        return self._rnday

    @rnday.setter
    def rnday(self, rnday: Union[float, timedelta, None]):
        if rnday is not None:
            if not isinstance(rnday, timedelta):
                rnday = timedelta(days=float(rnday))
            self._rnday = rnday / timedelta(days=1)
        else:
            self._rnday = None

    @property
    def dt(self) -> Union[float, None]:
        return self._dt

    @dt.setter
    def dt(self, dt: Union[float, timedelta, None]):
        if dt is None:
            dt = timedelta(seconds=150.0)

        if not isinstance(dt, timedelta):
            dt = timedelta(seconds=float(dt))

        self._dt = dt.total_seconds()

    @property
    def nspool(self):
        return self._nspool

    @nspool.setter
    def nspool(self, nspool: Union[int, float, timedelta, None]):
        if nspool is None and self.rnday is not None:
            nspool = int(round(self.rnday / self.dt))
        if isinstance(nspool, timedelta):
            nspool = int(round(nspool.total_seconds() / self.dt))
        if isinstance(nspool, float):
            nspool = int(round(timedelta(hours=nspool).total_seconds() / self.dt))
        if isinstance(nspool, (int, float)):
            if nspool < 0:
                raise ValueError("nspool must be positive.")
        if nspool is not None:
            self._nspool = int(nspool)
        else:
            self._nspool = None

    @property
    def ihfskip(self):
        return self._ihfskip

    @ihfskip.setter
    def ihfskip(self, ihfskip: Union[int, timedelta, None]):

        if not isinstance(ihfskip, (int, timedelta, type(None))):
            raise TypeError("Argument ihfskip must be int, timedelta or None.")

        if ihfskip is None and self.rnday is not None:
            ihfskip = int(round(timedelta(days=self.rnday).total_seconds() / self.dt))

        if isinstance(ihfskip, timedelta):
            ihfskip = int(round(ihfskip.total_seconds() / self.dt))

        if isinstance(self.nspool, int):
            if self.nspool > 0:
                if not (ihfskip / self.nspool).is_integer():
                    raise ValueError(
                        "ihfskip/nspool must be an integer but got "
                        "ihfskip/nspool="
                        f"{ihfskip}/{self.nspool}={ihfskip/self.nspool}"
                    )

        self._ihfskip = ihfskip

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, template: Union[str, os.PathLike, None]):
        if template is True:
            template = pathlib.Path(__file__) / 'param.nml'
        elif template is False or template is None:
            template = None
        else:
            template = pathlib.Path(template)
        self._template = template

    @property
    def defaults(self):
        if self.template is None:
            return f90nml.read(pathlib.Path(__file__).parent / 'param.nml')["core"]
        return f90nml.read(self.template)['core']



from datetime import timedelta
from typing import Union
import logging
import pathlib
import tempfile

import f90nml

# from pyschism.domain import ModelDomain
from pyschism.enums import Stratification
from pyschism.param.core import CORE
from pyschism.param.opt import OPT
from pyschism.param.schout import SCHOUT
from pyschism.param import schism_init
# from pyschism.stations import Stations


logger = logging.getLogger(__name__)


class Param:
    def __init__(
        self,
        dt: Union[int, float, timedelta] = None,
        rnday: Union[int, float, timedelta] = None,
        ibc: Union[Stratification, int, str] = Stratification.BAROTROPIC,
        nspool: Union[int, float, timedelta] = None,
        ihfskip: Union[int, timedelta] = None,
        template=None,
    ):

        self.core = CORE()
        self.core.ibc = ibc
        self.core.rnday = rnday
        self.core.dt = dt
        self.core.nspool = nspool
        self.core.ihfskip = ihfskip
        self.opt = OPT()
        self.schout = SCHOUT()
        self.template = template

    def __str__(self):
        return f"{str(self.core)}\n\n{str(self.opt)}\n\n{str(self.schout)}\n"

    def write(self, path, overwrite=False, use_template=False):
        path = pathlib.Path(path)
        if path.is_file() and not overwrite:
            raise IOError(f"File {path} exists and overwrite=False")
        if use_template:
            PARAM_TEMPLATE = pathlib.Path(__file__).parent / 'param.nml' if use_template is True else use_template
            schism_param_sample = schism_init.read_schism_param_sample_patched(PARAM_TEMPLATE)
            tmpfile = tempfile.NamedTemporaryFile()
            with open(tmpfile.name, 'w') as fh:
                fh.write(schism_param_sample)
            f90nml.patch(tmpfile.name, self.to_dict(), path)
        else:
            with open(path, "w") as f:
                f.write(str(self))

    def to_dict(self):
        return {
            "CORE": self.core.to_dict(),
            "OPT": self.opt.to_dict(),
            "SCHOUT": self.schout.to_dict(),
        }

    # @property
    # def core(self):
    #     return self._core

    # @property
    # def opt(self):
    #     return self._opt

    # @property
    # def schout(self):
    #     return self._schout

    # @property
    # def model_domain(self):
    #     return self.__model_domain

    # @property
    # def nhot_write(self):
    #     return self.schout.nhot_write

    # @property
    # def stations(self):
    #     return self.__stations

    # @property
    # def _model_domain(self):
    #     return self.__model_domain

    # @_model_domain.setter
    # def _model_domain(self, model_domain):
    #     assert isinstance(model_domain, ModelDomain), \
    #         f"Argument model_domain must be of type {ModelDomain}, " \
    #         f"not {type(model_domain)}."
    #     self.__model_domain = model_domain

    # @property
    # def _opt(self):
    #     return self.__opt

    # @_opt.setter
    # def _opt(self, opt: OPT):
    #     # friction parameters
    #     opt.nchi = self.model_domain.fgrid
    #     # set coordinate system
    #     opt.ics = self.model_domain.ics
    #     # set coriolis
    #     opt.ncor = self.model_domain.ncor
    #     # set atmospheric forcing
    #     if self.model_domain.nws is not None:
    #         opt.nws = self.model_domain.nws
    #     # TODO: Set the remaining options:
    #     # msc2
    #     # mdc2
    #     # ntracer_gen
    #     # ntracer_age
    #     # sed_class
    #     # eco_class
    #     self.__opt = opt

    # @property
    # def _nhot_write(self):
    #     return self.__nhot_write

    # @_nhot_write.setter
    # def _nhot_write(self, nhot_write: Union[int, timedelta, bool, None]):

    #     if not isinstance(nhot_write, (int, bool, timedelta, type(None))):
    #         raise TypeError(f"Argument nhot_write must be of type {int}, "
    #                         f"{bool}, {timedelta}, or None.")

    #     if nhot_write is True:
    #         nhot_write = int(round(self.core.rnday / self.core.dt))

    #     elif isinstance(nhot_write, timedelta):
    #         nhot_write = int(round(nhot_write / self.core.dt))

    #     if nhot_write is not None:
    #         if nhot_write % self.core.ihfskip != 0:
    #             raise ValueError("nhot_write must be a multiple of ihfskip")
    #         self.schout.nhot_write = nhot_write

    # @property
    # def _stations(self):
    #     return self.__stations

    # @_stations.setter
    # def _stations(self, stations: Union[Stations, None]):
    #     assert isinstance(stations, (Stations, type(None))), \
    #         f"Argument stations must be of type {Stations} or None, " \
    #         f"not type {type(stations)}."
    #     if isinstance(stations, Stations):
    #         stations.transform_to(self.model_domain.hgrid.crs)
    #         stations.clip(
    #             self.model_domain.hgrid.get_bbox()
    #         )
    #         self.schout.iout_sta = 1
    #         self.schout.nspool_sta = stations.nspool_sta
    #     self.__stations = stations



from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime, timedelta
import json
import logging
from multiprocessing import Pool, cpu_count
import os
import pathlib
import posixpath
import shutil

import tarfile
import tempfile
from time import time
from typing import Union
import urllib
import appdirs
import boto3
from botocore import UNSIGNED
from botocore.config import Config
import fiona
import geopandas as gpd
import matplotlib.pyplot as plt
from netCDF4 import Dataset
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
from shapely import ops
from shapely.geometry import LinearRing, Point, MultiPoint, LineString, box
import wget

from pyschism import dates
from pyschism.mesh.base import Gr3
from pyschism.utils.jsonencoder import NpEncoder

from pyschism.forcing.source_sink.base import SourceSink, Sources, Sinks

DATADIR = pathlib.Path(appdirs.user_data_dir("pyschism/nwm"))
DATADIR.mkdir(exist_ok=True, parents=True)

logger = logging.getLogger(__name__)

class NWMElementPairings:
    def __init__(self, hgrid: Gr3, nwm_file=None, workers=-1):

        # TODO: Accelerate using dask: https://blog.dask.org/2017/09/21/accelerating-geopandas-1

        self._nwm_file = nwm_file

        logger.info("Computing NWMElementPairings...")
        self._hgrid = hgrid

        # An STR-Index returns the reaches that are near the boundaries of the
        # mesh. This subsamples the NWM network, but also is not the exact
        # result. This is used to speed-up computations by filtering the input
        # data.
        logger.info("Computing r_index.")
        start = time()
        nwm_r_index = self.gdf.sindex
        logger.info(f"Computing r_index took {time() - start}.")

        # The r-index is used to find intersections between mesh boundary edges
        # and NWM reaches (approximate results)
        logger.info("Use r_index to filter features.")
        start = time()
        possible_indexes = set()
        for edge in hgrid.hull.edges().itertuples():
            for index in list(nwm_r_index.intersection(edge.geometry.bounds)):
                possible_indexes.add(index)
        possible_matches = self.gdf.iloc[list(possible_indexes)]
        logger.info(f"Filtering features took {time()-start}.")
        del possible_indexes
        del nwm_r_index

        # The hull rings itersections is used to find the exact NWM reaches
        # that intersect the mesh's hull.
        logger.info("Finding exact features intersections.")
        start = time()
        exact_indexes = set()
        for pm in possible_matches.itertuples():
            if hgrid.hull.rings().geometry.intersects(pm.geometry).any():
                exact_indexes.add(pm.Index)
        reaches = self.gdf.iloc[list(exact_indexes)]

        logger.info(f"Finding exact features took {time()-start}.")

        # release some memory
        del possible_matches
        del exact_indexes
        del self._gdf

        logger.info("Pairing features to corresponding element.")

        # Pair each reach with corresponding element.
        # 1) Find reach-hull intersections.
        start = time()
        data = []
        intersection: gpd.GeoDataFrame
        for i, reach in enumerate(reaches.itertuples()):
            for ring in hgrid.hull.rings().itertuples():
                if ring.geometry.intersects(reach.geometry):
                    intersections = ring.geometry.intersection(reach.geometry)
                    if isinstance(intersections, MultiPoint):
                        for point in intersections.geoms:
                            data.append({"geometry": point, "reachIndex": i})
                        continue

                    data.append({"geometry": intersections, "reachIndex": i})
                    break

        if len(data) == 0:
            # TODO: change for warning in future.
            raise IOError(
                "No National Water model intersections found on the mesh.")
        intersection = gpd.GeoDataFrame(data, crs=hgrid.crs)
        #TODO: add exporting intersection as an option
        #intersection.to_file('intersections.shp')
        del data

        # 2) Generate element centroid KDTree
        centroids = []
        for element in hgrid.elements.elements.values():
            cent = LinearRing(
                hgrid.nodes.coord[list(
                    map(hgrid.nodes.get_index_by_id, element))]
            ).centroid
            centroids.append((cent.x, cent.y))
        tree = cKDTree(centroids)
        del centroids

        # 3) Match reach/boundary intersection to nearest element centroid
        coords = [
            np.array(inters.geometry.coords) for inters in intersection.itertuples()
        ]
        _, idxs = tree.query(np.vstack(coords), workers=workers)
        del tree

        logger.info(
            "Pairing features to corresponding element took " f"{time()-start}."
        )

        hull = hgrid.hull.multipolygon()

        start = time()
        sources = defaultdict(list)
        sinks = defaultdict(list)
        for row in intersection.itertuples():
            poi = row.geometry
            reach = reaches.iloc[row.reachIndex].geometry
            if not isinstance(reach, LineString):
                reach = ops.linemerge(reach)
            for segment in map(
                LineString, zip(reach.coords[:-1], reach.coords[1:])
            ):
                if segment.intersects(poi.buffer(np.finfo(np.float32).eps)):
                    segment_origin = Point(segment.coords[0])
                    d1 = segment_origin.distance(poi)
                    downstream = segment.interpolate(
                        d1 + np.finfo(np.float32).eps)
                    element = hgrid.elements.gdf.iloc[idxs[row.Index]]
                    if (
                        box(*LineString([poi, downstream]).bounds)
                        .intersection(hull)
                        .intersects(downstream)
                    ):
                        sources[element.id].append(reaches.iloc[row.reachIndex].feature_id)
                    else:
                        sinks[element.id].append(reaches.iloc[row.reachIndex].feature_id)
                    break

        logger.info(
            "Sorting features into sources and sinks took: " f"{time()-start}.")
        self.sources = sources
        self.sinks = sinks

    def make_plot(self):
        # verification plot
        data = []
        egdf = self.hgrid.elements.gdf
        for eid in self.sources.keys():
            eidx = self.hgrid.elements.get_index_by_id(eid)
            data.append({"geometry": egdf.iloc[eidx].geometry})
        src_gdf = gpd.GeoDataFrame(data)
        data = []
        for eid in self.sinks.keys():
            eidx = self._hgrid.elements.get_index_by_id(eid)
            data.append({"geometry": egdf.iloc[eidx].geometry})
        snk_gdf = gpd.GeoDataFrame(data)

        ax = egdf.plot(facecolor="none", edgecolor="black", lw=0.7)
        src_gdf.plot(color="red", ax=ax, alpha=0.5)
        snk_gdf.plot(color="blue", ax=ax, alpha=0.5)
        plt.show()

    def save_json(self, sources=True, sinks=True):

        sources = "sources.json" if sources is True else sources
        if sources:
            logger.info(f"Saving {sources}")
            with open(sources, "w") as fh:
                json.dump(self.sources, fh, cls=NpEncoder)

        sinks = "sinks.json" if sinks is True else sinks
        if sinks:
            with open(sinks, "w") as fh:
                logger.info(f"Saving {sinks}")
                json.dump(self.sinks, fh, cls=NpEncoder)

    @staticmethod
    def load_json(hgrid, sources=None, sinks=None):
        pairings = NWMElementPairings.__new__(NWMElementPairings)
        logger.info(f"Loading pairing sources: {sources}")
        pairings.sources = json.load(
            open(sources)) if sources is not None else {}
        logger.info(f"Loading pairing sinks: {sinks}")
        pairings.sinks = json.load(open(sinks)) if sinks is not None else {}
        pairings._hgrid = hgrid
        return pairings

    @property
    def sources_gdf(self):
        if not hasattr(self, "_sources_gdf"):
            data = []
            for eid, features in self.sources.items():
                eidx = self.hgrid.elements.get_index_by_id(eid)
                data.append(
                    {
                        "element_id": eid,
                        "geometry": LineString(
                            self.hgrid.elements.gdf.loc[eidx].geometry.exterior.coords
                        ),
                        "features": " ".join(list(map(str, features))),
                    }
                )
            self._sources_gdf = gpd.GeoDataFrame(data)
        return self._sources_gdf

    @property
    def sinks_gdf(self):
        if not hasattr(self, "_sinks_gdf"):
            data = []
            for eid, features in self.sinks.items():
                eidx = self.hgrid.elements.get_index_by_id(eid)
                data.append(
                    {
                        "element_id": eid,
                        "geometry": LineString(
                            self.hgrid.elements.gdf.loc[eidx].geometry.exterior.coords
                        ),
                        "features": " ".join(list(map(str, features))),
                    }
                )
            self._sinks_gdf = gpd.GeoDataFrame(data)
        return self._sinks_gdf

    @property
    def hgrid(self):
        return self._hgrid

    @property
    def gdf(self):
        if not hasattr(self, "_gdf"):
            gdf_coll = []
            #for reach_layer in [
            #    reach_layer
            #    for reach_layer in fiona.listlayers(self.nwm_file)
            #    if "reaches" in reach_layer
            #]:
            reach_layers = ['nwm_reaches_conus']
            for reach_layer in reach_layers:
                logger.info(f'layer is {reach_layer}')
                layer_crs = gpd.read_file(
                    self.nwm_file, rows=1, layer=reach_layer).crs
                bbox = self.hgrid.get_bbox(crs=layer_crs)
                gdf_coll.append(
                    gpd.read_file(
                        self.nwm_file,
                        bbox=(bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax),
                        layer=reach_layer,
                    )
                )
            #self._gdf = pd.concat(gdf_coll)
            self._gdf = gpd.GeoDataFrame(gdf_coll[0])
        return self._gdf

    @property
    def nwm_file(self):
        return self._nwm_file

    @property
    def _nwm_file(self):
        return self.__nwm_file

    @_nwm_file.setter
    def _nwm_file(self, nwm_file):
        nwm_file = (
            list(DATADIR.glob("**/*hydrofabric*.gdb")
                 ) if nwm_file is None else nwm_file
        )
        if isinstance(nwm_file, list):
            if len(nwm_file) == 0:
                tmpdir = tempfile.TemporaryDirectory()
                logger.info(
                    f"Downloading National Water Model stream network tar file to {tmpdir}"
                )
                try:
                    wget.download(
                        "https://www.nohrsc.noaa.gov/pub/staff/keicher/NWM_live/web/data_tools/NWM_channel_hydrofabric.tar.gz",
                        out=tmpdir.name,
                        bar=wget.bar_adaptive
                        if logger.getEffectiveLevel() < 30
                        else None,
                    )
                except urllib.error.HTTPError as e:
                    logger.fatal(
                        "Could not download NWM_channel_hydrofabric.tar.gz")
                    raise e
                tmpfile = list(pathlib.Path(
                    tmpdir.name).glob("**/*.tar.gz"))[0]
                with tarfile.open(tmpfile, "r:gz") as src:
                    logger.info(
                        f"Extracting National Water Model stream network tar file to {DATADIR}"
                    )

                    def is_within_directory(directory, target):
                        
                        abs_directory = os.path.abspath(directory)
                        abs_target = os.path.abspath(target)
                    
                        prefix = os.path.commonprefix([abs_directory, abs_target])
                        
                        return prefix == abs_directory
                    
                    def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                    
                        for member in tar.getmembers():
                            member_path = os.path.join(path, member.name)
                            if not is_within_directory(path, member_path):
                                raise Exception("Attempted Path Traversal in Tar File")
                    
                        tar.extractall(path, members, numeric_owner=numeric_owner)
                        
                    
                    safe_extract(src, DATADIR)
                nwm_file = list(DATADIR.glob("**/*.gdb"))[0]
            elif len(nwm_file) == 1:
                nwm_file = nwm_file[0]
            else:
                raise Exception("Found more than 1 NWM hydrofabric file.")
        self.__nwm_file = nwm_file

def get_aggregated_features(nc_feature_id, features):
    aggregated_features = []
    for source_feats in features:
        aggregated_features.extend(list(source_feats))

    in_file=[]
    for feature in aggregated_features:
        idx=np.where(nc_feature_id == int(feature))[0]
        in_file.append(idx.item())

    in_file_2 = []
    sidx = 0
    for source_feats in features:
        eidx = sidx + len(source_feats)
        #in_file_2.append(in_file[sidx:eidx].tolist())
        in_file_2.append(in_file[sidx:eidx])
        sidx = eidx
    return in_file_2

def streamflow_lookup(file, indexes, threshold=-1e-5):
    nc = Dataset(file)
    streamflow = nc["streamflow"][:]
    streamflow[np.where(streamflow < threshold)] = 0.0
    #change masked value to zero
    streamflow[np.where(streamflow.mask)] = 0.0
    data = []
    for indxs in indexes:
        # Note: Dataset already consideres scale factor and offset.
        data.append(np.sum(streamflow[indxs]))
    nc.close()
    return data

class AWSDataInventory(ABC):
    def __new__(
        cls, start_date, rnday, product=None, verbose=False, fallback=True, cache=None
    ):
        # AWSHindcastInventory
        # The latest(as of 12/21/2021) AWSHindcast dataset covers from Feb 1979 through Dec 2020
        if start_date >= dates.localize_datetime(
            datetime(1979, 2, 1, 0, 0)
        ) and start_date + rnday <= dates.localize_datetime(
            datetime(2020, 12, 31, 23, 59)
        ):
            return AWSHindcastInventory.__new__(cls)

        # GOOGLEHindcastInventory -> January 2019 through 30 days earlier than today
        # data before April 20, 2021 (including Apirl 20) is 3-hr interval, after that is hourly
        # Missing data 20211231, 20220101, 20220102, 20220103
        elif start_date >= dates.localize_datetime(
            datetime(2021, 1, 1, 0, 0)
        ) and start_date + rnday < dates.nearest_zulu() - timedelta(days=2):
            return GOOGLEHindcastInventory.__new__(cls)

        elif start_date >= dates.nearest_zulu() - timedelta(days=2):
            return AWSForecastInventory.__new__(cls)

        else:
            raise Exception(
                f"No NWM model data for start_date {start_date} and end_date {start_date+rnday}."
            )

    #@abstractmethod
    #def request_data(self, request_time):
    #    raise NotImplementedError

    #@property
    #@abstractmethod
    #def bucket(self):
    #    raise NotImplementedError

    @property
    def nearest_cycle(self) -> datetime:
        return dates.nearest_cycle(self.start_date)

    #@property
    #def s3(self):
    #    try:
    #        return self._s3
    #    except AttributeError:
    #        self._s3 = boto3.client(
    #            "s3", config=Config(signature_version=UNSIGNED))
    #        return self._s3

    @property
    def tmpdir(self):
        if not hasattr(self, "_tmpdir"):
            self.__tmpdir = tempfile.TemporaryDirectory()
            self._tmpdir = pathlib.Path(self.__tmpdir.name)
        return self._tmpdir

    @property
    def files(self):
        return self._files


class AWSHindcastInventory(AWSDataInventory):
    def __new__(cls):
        return object.__new__(AWSHindcastInventory)

    def __init__(
        self,
        start_date: datetime = None,
        rnday: Union[int, float, timedelta] = timedelta(days=5.0),
        product=None,
        verbose=False,
        fallback=True,
        cache=None,
    ):
        """This will download the National Water Model retro data.
        A 42-year (February 1979 through December 2020) retrospective
        simulation using version 2.1 of the NWM.
        """
        self.product = "CHRTOUT_DOMAIN1.comp" if product is None else product
        self.cache = cache
        self.start_date = (
            dates.nearest_cycle()
            if start_date is None
            else dates.nearest_cycle(dates.localize_datetime(start_date))
        )
        # self.start_date = self.start_date.replace(tzinfo=None)
        self.rnday = rnday if isinstance(
            rnday, timedelta) else timedelta(days=rnday)
        self.fallback = fallback
        self._files = {
            _: None
            for _ in np.arange(
                self.start_date,
                self.start_date + self.rnday + self.output_interval,
                self.output_interval,
            ).astype(datetime)
        }

        end_date = self.start_date + self.rnday

        years = np.arange(self.start_date.year, end_date.year+1)
        
        file_metadata = [] 
        for it, year in enumerate(years):
            paginator = self.s3.get_paginator("list_objects_v2")
            pages = paginator.paginate(
                Bucket=self.bucket, Prefix=f"model_output/{year}"
            )

            self.data = []
            for page in pages:
                for obj in page["Contents"]:
                    self.data.append(obj)

            metadata = sorted([_["Key"] for _ in self.data if "CHRTOUT_DOMAIN1.comp" in _["Key"]])
            [file_metadata.append(i) for i in metadata] 

        timevector = np.arange(
            datetime(self.start_date.year, 1, 1),
            datetime(end_date.year + 1, 1, 1),
            np.timedelta64(1, "h"),
            dtype="datetime64",
        )

        timefile = {
            pd.to_datetime(str(timevector[i])): file_metadata[i]
            for i in range(len(timevector))
        }

        for requested_time in self._files:
            logger.info(f"Requesting NWM data for time {requested_time}")
            self._files[requested_time] = self.request_data(
                timefile.get(requested_time)
            )

    def request_data(self, key):
        filename = self.tmpdir / key
        if filename.is_file() is False:
            cached_file = list(self.tmpdir.glob(f'**/{filename.name}'))
            if len(cached_file) == 1:
                filename = cached_file[0]
                logger.info(f"Using cached file {filename}, ")
            else:
                filename.parent.mkdir(parents=True, exist_ok=True)
                tmpfile = tempfile.NamedTemporaryFile().name
                with open(tmpfile, "wb") as f:
                    logger.info(f"Downloading file {key}, ")
                    self.s3.download_fileobj(self.bucket, key, f)
                shutil.move(tmpfile, filename)
        return filename

    @property
    def bucket(self):
        #return "noaa-nwm-retro-v2.0-pds"
        return "noaa-nwm-retrospective-2-1-pds"

    @property
    def s3(self):
        try:
            return self._s3
        except AttributeError:
            self._s3 = boto3.client(
                "s3", config=Config(signature_version=UNSIGNED))
            return self._s3

    @property
    def output_interval(self) -> timedelta:
        return {"CHRTOUT_DOMAIN1.comp": timedelta(hours=1)}[self.product]

    @property
    def cached_files(self):
        return sorted(list(self.tmpdir.glob("**/*.comp")))

    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, cache: Union[str, os.PathLike, None, bool]):
        if cache is None or cache is False:
            self._cache = False
        elif cache is True:
            self._cache = pathlib.Path(
                appdirs.user_cache_dir(
                    f"pyschism/nwm/hindcast_data/{self.product}")
            )
            self._cache.mkdir(exist_ok=True, parents=True)
            self._tmpdir = self._cache

        elif isinstance(cache, (str, os.PathLike)):
            self._cache = pathlib.Path(cache)
            self._cache.mkdir(exist_ok=True, parents=True)
            self._tmpdir = self._cache
        else:
            raise TypeError(
                f"Unhandled argument cache={cache} of type {type(cache)}.")

class GOOGLEHindcastInventory(AWSDataInventory):
    def __new__(cls):
        return object.__new__(GOOGLEHindcastInventory)

    def __init__(
        self,
        start_date: datetime = None,
        rnday: Union[int, float, timedelta] = timedelta(days=5.0),
        product='medium_range_mem1',
        verbose=False,
        fallback=True,
        cache=None,
    ):

        self.start_date = dates.nearest_cycle() if start_date is None \
            else dates.nearest_cycle(dates.localize_datetime(start_date))
        self.rnday = rnday if isinstance(rnday, timedelta) \
            else timedelta(days=rnday)
        self.product = product
        self.verbose = verbose
        self.fallback = fallback
        self.cache = cache
        self._files = {_: None for _ in np.arange(
            self.start_date,
            self.start_date + self.rnday + self.output_interval,
            self.output_interval
        ).astype(datetime)}

        timevector=np.arange(datetime(self.start_date.year, 1, 1),
            datetime(self.start_date.year+1,1, 1),
            np.timedelta64(1, 'h'),
            dtype='datetime64')

        for requested_time, _ in self._files.items():

            logger.info(f'Requesting NWM data for time {requested_time}')

            if requested_time.hour == 0:
                requested_time2 = requested_time - timedelta(days=1)
            else:
                requested_time2 = requested_time

            self._files[requested_time] = self.request_data(requested_time, requested_time2)

    def request_data(self, request_time, request_time2):

        fname = self.tmpdir / f'nwm.t00z.{self.product[:12]}.channel_rt_1.' \
            f'{request_time.strftime("%Y%m%d%H")}.conus.nc'
        #logger.info(f'fname is {fname}')

        if fname.is_file():
            cached_file = list(self.tmpdir.glob(f'**/{fname.name}'))
            if len(cached_file) == 1:
                fname = cached_file[0]
                logger.info(f'Use cached file {fname}')
        else:
        
            fname = f'{self.start_date.strftime("%Y%m%d")}/nwm.t00z.' \
                f'{self.product[:12]}.channel_rt_1.{request_time.strftime("%Y%m%d%H")}.conus.nc'

            logger.info(f'Downloading file {request_time}, ')

            it = request_time.strftime("%H")
            if it == '00':
                logger.info(f'Requesting data at 00Z from yesterday!')
                it = str(int(it) + 24)
            it = it.zfill(3)

            url = f'https://storage.googleapis.com/national-water-model/nwm.{request_time2.strftime("%Y%m%d")}' \
                f'/{self.product}/nwm.t00z.{self.product[:12]}.channel_rt_1.f{it}.conus.nc'
            logger.info(f'{url}')
            try:
                wget.download(url, fname)
            except:
                logger.info(f'No data for {request_time}!')
                
        return fname

    @property
    def output_interval(self) -> timedelta:
        return {
            'medium_range_mem1': timedelta(hours=1)
        }[self.product]

    @property
    def cached_files(self):
        return sorted(list(self.tmpdir.glob("**/*.comp")))

    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, cache: Union[str, os.PathLike, None, bool]):
        if cache is None or cache is False:
            self._cache = False
        elif cache is True:
            self._cache = pathlib.Path(
                appdirs.user_cache_dir(
                    f"pyschism/nwm/hindcast_data/{self.product}")
            )
            self._cache.mkdir(exist_ok=True, parents=True)
            self._tmpdir = self._cache

        elif isinstance(cache, (str, os.PathLike)):
            self._cache = pathlib.Path(cache)
            self._cache.mkdir(exist_ok=True, parents=True)
            self._tmpdir = self._cache
        else:
            raise TypeError(
                f"Unhandled argument cache={cache} of type {type(cache)}.")

class AWSForecastInventory(AWSDataInventory):
    def __new__(cls):
        return object.__new__(AWSForecastInventory)

    def __init__(
        self,
        start_date: datetime = None,
        rnday: Union[int, float, timedelta] = timedelta(days=5.0),
        product=None,
        verbose=False,
        fallback=True,
        cache=None,
    ):
        """This will download the latest National Water Model data.

        NetCDF files are saved to the system's temporary directory.
        The AWS data goes back 30 days. For requesting hindcast data from
        before we need a different data source
        """
        self.product = "medium_range_mem1" if product is None else product
        self.cache = cache
        self.start_date = (
            dates.nearest_cycle()
            if start_date is None
            else dates.nearest_cycle(dates.localize_datetime(start_date))
        )

        yesterday = self.start_date - timedelta(days=1)
       
        #check if previous day's data folder exists
        if not os.path.exists(yesterday.strftime("%Y%m%d")):
            logger.info(f"Downloading NWM data for {yesterday}")
            _ = self.download(yesterday, days=1)
        
        filemaps = self.download(self.start_date, days=rnday.days)

        # self.start_date = self.start_date.replace(tzinfo=None)
        self.rnday = rnday if isinstance(
            rnday, timedelta) else timedelta(days=rnday)
        self.fallback = fallback
        self._files = {
            _: None
            for _ in np.arange(
                self.start_date, # + timedelta(hours=3),
                self.start_date + self.rnday + self.output_interval,
                self.output_interval,
            ).astype(datetime)
        }

        for it, (requested_time, _) in enumerate(self._files.items()):

            logger.info(f"Requesting NWM data for time {requested_time}")
            if it == 0 and requested_time.hour == 0:
                f0 = f'{yesterday.strftime("%Y%m%d")}/nwm.' \
                    + yesterday.strftime("%Y%m%d") \
                    + '/medium_range_mem1/nwm.t00z.medium_range.channel_rt_1.f024.conus.nc'
                logger.info(f'Using data {f0}')
                self.files[requested_time] = f0
                continue

            logger.info(f"Using data {filemaps[requested_time]}")
            self._files[requested_time] = filemaps[requested_time]

    def download(self, nwmdate, days):
        self.data = self.s3.list_objects_v2(
            Bucket=self.bucket,
            Delimiter="/",
            Prefix=f'nwm.{nwmdate.strftime("%Y%m%d")}' f"/{self.product}/",
        )

        file_metadata = list(
                sorted(
                    [
                        _["Key"]
                        for _ in self.data["Contents"]
                        if "channel" in _["Key"] and "Contents" in self.data
                    ]
                )
            )

        filedir = pathlib.Path(nwmdate.strftime("%Y%m%d"))
        filedir.mkdir(exist_ok=True, parents=True)

        filedict = {}
        for key in file_metadata[0:days*24+1]:
            filetime = self.key2date(key)
            filename = nwmdate.strftime("%Y%m%d") + '/' + key
            filesubdir = pathlib.Path(filename)
            filesubdir.parent.mkdir(parents=True, exist_ok=True)

            logger.info(f"Downloading file {key}, ")
            self.s3.download_file(self.bucket, key, filename)
            filedict[filetime] = filename
       
        return filedict

    def key2date(self, key):
        base_date_str = f'{key.split("/")[0].split(".")[-1]}'
        timedelta_str = key.split("channel_rt_1.")[-1].split(".")[0].strip("f")
        return (
            datetime.strptime(base_date_str, "%Y%m%d")
            + timedelta(hours=float(timedelta_str))
            # + timedelta(hours=int(key.split('.')[2].strip('tz')))
            # + timedelta(hours=float(timedelta_str))
        )

    @property
    def bucket(self):
        return "noaa-nwm-pds"
    
    @property
    def s3(self):
        try:
            return self._s3
        except AttributeError:
            self._s3 = boto3.client(
                "s3", config=Config(signature_version=UNSIGNED))
            return self._s3

    @property
    def output_interval(self) -> timedelta:
        return {"medium_range_mem1": timedelta(hours=1)}[self.product]

    @property
    def timevector(self):
        return np.arange(
            self.start_date,
            self.start_date + self.rnday + self.output_interval,
            self.output_interval,
        ).astype(datetime)

    #@property
    #def files(self):
    #    return sorted(list(self.tmpdir.glob("**/*.nc")))

    @property
    def requested_product(self):
        return {"medium_range_mem1": "medium_range.channel_rt_1"}[self.product]

    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, cache: Union[str, os.PathLike, None, bool]):
        if cache is None or cache is False:
            self._cache = False
        elif cache is True:
            self._cache = pathlib.Path(
                appdirs.user_cache_dir(
                    f"pyschism/nwm/forecast_data/{self.product}")
            )
            self._cache.mkdir(exist_ok=True, parents=True)
            self._tmpdir = self._cache

        elif isinstance(cache, (str, os.PathLike)):
            self._cache = pathlib.Path(cache)
            self._cache.mkdir(exist_ok=True, parents=True)
            self._tmpdir = self._cache
        else:
            raise TypeError(
                f"Unhandled argument cache={cache} of type {type(cache)}.")


class NationalWaterModel(SourceSink):

    start_date = dates.StartDate()
    end_date = dates.EndDate()
    # run_days = dates.RunDays()

    def __init__(
        self, aggregation_radius=None, pairings=None, nwm_file=None, cache=False
    ):
        self.nwm_file = nwm_file
        self.aggregation_radius = aggregation_radius
        self.pairings = pairings
        self.cache = cache
        self._data = None

    def _fetch_data(
        self,
        gr3: Gr3,
        start_date: datetime = None,
        end_date: Union[datetime, timedelta] = None,
        nprocs=-1,
        product=None,
    ):
        nprocs = -1 if nprocs is None else nprocs
        nprocs = cpu_count() if nprocs == -1 else nprocs
        self.start_date = start_date
        self.end_date = end_date
        self.pairings = (
            NWMElementPairings(gr3, nwm_file=self.nwm_file, workers=nprocs)
            if self.pairings is None
            else self.pairings
        )
        self._inventory = AWSDataInventory(
            start_date=self.start_date,
            rnday=self.end_date - self.start_date,
            cache=self.cache,
            # product="medium_range_mem1",
        )

        self._timevector = [
            dates.localize_datetime(d) for d in self.inventory.files.keys()
        ]

        #src_idxs, snk_idxs = self.inventory.get_nc_pairing_indexes(
        #    self.pairings)
        #logger.info(f'Start aggregating NWM timeseries using nprocs={nprocs}')
        start0 = datetime.now()
        #with Pool(processes=nprocs) as pool:
        #    sources = pool.starmap(
        #        streamflow_lookup,
        #        [(file, self.pairings.sources) for file in self.inventory.files.values()],
        #    )
        #    sinks = pool.starmap(
        #        streamflow_lookup,
        #        [(file, self.pairings.sinks) for file in self.inventory.files.values()],
        #    )
        #pool.join()

        sources = []
        sinks = []
        nc_fid0 = Dataset(list(self.inventory.files.values())[0])["feature_id"][:]
        src_idxs = get_aggregated_features(nc_fid0, self.pairings.sources.values())
        snk_idxs = get_aggregated_features(nc_fid0, self.pairings.sinks.values())
        for file in self.inventory.files.values():
            start = datetime.now()
            nc = Dataset(file)
            ncfeatureid=nc['feature_id'][:]
            if not np.all(ncfeatureid == nc_fid0):
                logger.info(f'Indexes of feature_id are changed in  {file}')
                src_idxs=get_aggregated_features(ncfeatureid, self.pairings.sources.values())
                snk_idxs=get_aggregated_features(ncfeatureid, self.pairings.sinks.values())
                nc_fid0 = ncfeatureid

            sources.append(streamflow_lookup(file, src_idxs))
            sinks.append(streamflow_lookup(file, snk_idxs))
            logger.info(f'Processing file {file} took {datetime.now() - start}')
            nc.close()

        source_data = {}
        sink_data = {}
        for i, file in enumerate(self.inventory.files.values()):
            nc = Dataset(file)
            _time = dates.localize_datetime(
                datetime.strptime(nc.model_output_valid_time,
                                  "%Y-%m-%d_%H:%M:%S")
            )
            for j, element_id in enumerate(self.pairings.sources):
                source_data.setdefault(_time, {})[element_id] = {
                    "flow": sources[i][j],
                    "temperature": -9999.0,
                    "salinity": 0.0,
                }

            for k, element_id in enumerate(self.pairings.sinks):
                sink_data.setdefault(_time, {})[element_id] = {
                    "flow": -sinks[i][k],
                }
            nc.close()
        logger.info(f'Timeseries aggregation took {datetime.now() - start0}')
        self._sources = Sources(source_data)
        self._sinks = Sinks(sink_data)
        self._data = {**source_data, **sink_data}

    def write(
        self,
        output_directory,
        gr3: Gr3,
        start_date: datetime = None,
        end_date: Union[datetime, timedelta] = None,
        overwrite: bool = False,
        nprocs=-1,
        product=None,
        msource: Union[str, bool] = True,
        vsource: Union[str, bool] = True,
        vsink: Union[str, bool] = True,
        source_sink: Union[str, bool] = True,
    ):
        if self._data is None:
            self._data = {}
            self._fetch_data(
                gr3,
                start_date=start_date,
                end_date=end_date,
                nprocs=nprocs,
                product=product,
            )

        if self.aggregation_radius is not None:
            self.aggregate_by_radius(gr3, self.aggregation_radius)

        super().write(
            output_directory,
            overwrite=overwrite,
            msource=msource,
            vsource=vsource,
            vsink=vsink,
            source_sink=source_sink,
        )

    @property
    def timevector(self):
        return self._timevector

    @property
    def inventory(self):
        return self._inventory



import logging
import os
import pathlib
import glob

import appdirs
from netCDF4 import Dataset
import numpy as np
from scipy.interpolate import griddata
from scipy.interpolate.fitpack2 import RectBivariateSpline

from pyschism.forcing.bctides.base import TidalDataProvider


logger = logging.getLogger(__name__)

FES2014_TIDES_EXTRA = 'ocean_tide_extrapolated'
FES2014_EASTWARD_VEL = 'eastward_velocity'
FES2014_NORTHWARD_VEL = 'northward_velocity'


def raise_missing_file(fpath, fname):
    raise FileNotFoundError('\n'.join([
        f'No FES2014 file found at "{fpath}".',
        'New users will need to register and download a copy of '
        f'the FES2014 NetCDF file (specifically `{fname}`) '
        'from https://www.aviso.altimetry.fr/en/data/products/auxiliary-products/global-tide-fes.html',
        'Once you obtain netCDF files, you can follow one of the '
        'following options: ',
        f'1) copy or symlink the file to "{fpath}"',
        f'2) set the environment variable `{fname}` to point'
        ' to the file',
    ]))


class FES2014(TidalDataProvider):

    def __init__(self, resource=None):
        self.resource = resource

    def get_elevation(self, constituent, vertices):
        logger.info('Querying FES2014 for elevation constituent '
                    f'{constituent}.')
        amp = self._get_interpolation(
            'elevation', 'amplitude', constituent, vertices)
        phase = self._get_interpolation(
            'elevation', 'phase', constituent, vertices)
        return amp, phase

    def get_velocity(self, constituent, vertices):
        logger.info('Querying FES2014 for velocity constituent '
                    f'{constituent}.')
        uamp = self._get_interpolation(
            'eastward_vel', 'Ua', constituent, vertices)
        uphase = self._get_interpolation(
            'eastward_vel', 'Ug', constituent, vertices)
        vamp = self._get_interpolation(
            'northward_vel', 'Va', constituent, vertices)
        vphase = self._get_interpolation(
            'northward_vel', 'Vg', constituent, vertices)
        return uamp, uphase, vamp, vphase

    @property
    def constituents(self):
        return ['M2', 'S2', 'N2', 'K2', 'K1', 'O1', 'P1',
                'Q1', 'Mm', 'Mf', 'M4', 'MN4', 'MS4', '2N2', 'S1']
        if not hasattr(self, '_constituents'):
            files = glob.glob(appdirs.user_data_dir('fes2014') + '/ocean_tide_extrapolated/*.nc')
            self._constituents = [f.split('/')[-1].split('.')[0] for f in files]
        return self._constituents

    def _get_resource(self, variable, constituent) -> Dataset:
        resource = self._resource[variable][constituent]
        if resource is not None:
            return Dataset(resource)
        datapath = pathlib.Path(appdirs.user_data_dir('fes2014'))
        if variable == 'elevation':
            fname = datapath / f'{FES2014_TIDES_EXTRA}/{constituent.lower()}.nc'
        if variable == 'eastward_vel':
            fname = datapath / f'{FES2014_EASTWARD_VEL}/{constituent.lower()}.nc'
        if variable == 'northward_vel':
            fname = datapath / f'{FES2014_NORTHWARD_VEL}/{constituent.lower()}.nc'
        return Dataset(fname)

    def _get_interpolation(self, phys_var, ncvar, constituent, vertices):
        ds = self._get_resource(phys_var, constituent) 
        lon = ds['lon'][:]
        lat = ds['lat'][:]
        dxs = np.unique(np.diff(lon))
        dys = np.unique(np.diff(lat))
        if len(dxs) !=1 or len(dys) != 1:
           raise ValueError(f'{phys_var}: lon, lat of {constituent}.nc not uniform! ')  
        dx = dxs[0]
        dy = dys[0]

        #get interp index
        xi = np.asarray(
            [x + 360. if x < 0. else x for x in vertices[:, 0]]).flatten()
        yi = vertices[:, 1].flatten()
        idx = np.floor((xi - lon[0]) / dx).astype('int')
        mask = np.nonzero((lon[idx] - xi) > 0)[0]
        idx[mask] = idx[mask] - 1
        idy = np.floor((yi - lat[0]) / dy).astype('int')
        mask = np.nonzero((lat[idy] - yi) > 0)[0]
        idy[mask] = idy[mask] - 1
        xrat = (xi - lon[idx]) / (lon[idx+1] - lon[idx])
        yrat = (yi - lat[idy]) / (lat[idy+1] - lat[idy])
        if np.sum((xrat > 1) | (xrat < 0) | (yrat > 1) | (yrat < 0)) != 0:
            raise ValueError(f'xrat or yrat > 1 or < 0')
       
        zi = ds[ncvar][:,:]
        # vm = 100 junk for amplitude, vm = 370 junk for phase
        if ncvar == 'amplitude' or ncvar == 'Ua' or ncvar == 'Va':
            vm = 100
            zi = zi / 100
        if ncvar == 'phase' or ncvar == 'Ug' or ncvar == 'Vg':
            vm = 370
            #Put phases into [0, 360)
            mask = zi<0
            zi[mask] = zi[mask] + 360
        v0 = np.c_[zi[idy, idx], zi[idy, idx+1], zi[idy+1, idx], zi[idy+1, idx+1]].T
        vmax = v0.max(axis=0)
        vmin = v0.min(axis=0)
        #handle phase jump
        if ncvar == 'phase' or ncvar == 'Ug' or ncvar == 'Vg':
            for kk in np.nonzero(abs(vmax - vmin) > 180)[0]:
                mask = abs(v0[:, kk] - vmax[kk]) > 180
                v0[mask, kk] = v0[mask, kk] + 360
        v1 = v0[0] * (1 - xrat) + v0[1] * xrat
        v2 = v0[2] * (1 - xrat) + v0[3] * xrat
        values = v1 * (1 - yrat) + v2 * yrat
        mask = np.nonzero((vmax > vm) * (vmin <= vm) * (vmin >= 0))[0]
        values[mask] = vmin[mask]
        if sum((vmax > vm) * ((vmin > vm) | (vmin < 0))) != 0:
            raise ValueError('All junk values for {phys_var} {constituent}')
        
        ds.close()
        return values

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, resource):
        if resource is None:
            resource = {'elevation': {}, 'eastward_vel': {}, 'northward_vel': {}}
            for constituent in self.constituents:
                for key in resource.keys():
                    resource[key].update({constituent: None})
        else:
            raise NotImplementedError('Check that static files exist.')
            for file in pathlib.Path(resource).glob('*.nc'):
                print(file)
        self._resource = resource



from datetime import datetime
import logging
from time import time

from matplotlib.transforms import Bbox

from pyschism.forcing.hycom.hycom2schism import DownloadHycom
from pyschism.mesh.hgrid import Hgrid

'''
Download hycom data for Fortran scripts.
Default is to download data for generating initial condition (use hgrid 
    as parameter to cover the whole domain).
Optionally, download data for generating th.nc (bnd=True) and nu.nc (nudge=True) 
    (use bbox as parameter to cut small domain).
'''
logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logger = logging.getLogger('pyschism')
logger.setLevel(logging.INFO)

if __name__ == '__main__':
    hgrid = Hgrid.open('./hgrid.gr3', crs='epsg:4326')
    date=datetime(2018, 1, 1)

    #xmin = -65
    #xmax = -50
    #ymin = 0
    #ymax = 53
    #bbox = Bbox.from_extents(xmin, ymin, xmax, ymax)
    hycom = DownloadHycom(hgrid=hgrid)
    t0 = time()
    hycom.fetch_data(date, rnday=1, bnd=False, nudge=False, sub_sample=3, outdir='./')
    print(f'It took {(time()-t0)/60} mins to download')
    



import pathlib
from typing import List, Tuple, Union

from matplotlib.collections import PolyCollection
import numpy as np
from shapely.geometry import Polygon, MultiPolygon, LinearRing

from pyschism.mesh.base import Gr3
from pyschism.figures import figure


class Prop:

    def __init__(self, gr3: Gr3, element_values: np.array):

        if not isinstance(gr3, Gr3):
            raise TypeError(
                f'Argument gr3 must be an instance of type {Gr3}, not type '
                f'{type(gr3)}.')

        values = np.array(element_values).flatten()
        if len(gr3.elements) != values.shape[0]:
            raise ValueError(
                'Shape mismatch between element_values and hgrid.')

        self.gr3 = gr3
        self.values = values

    def __str__(self):
        f = []
        for i, (iele, element) in enumerate(self.gr3.elements.to_dict().items()):
            f.append(f'{iele} {self.values[i]:G}')
        return '\n'.join(f)

    @classmethod
    def constant(cls, gr3: Gr3, value: np.array):
        return cls(gr3, np.full((len(gr3.elements),), value))

    def write(self, path, overwrite: bool = False):
        path = pathlib.Path(path)
        if path.exists() and not overwrite:
            raise IOError('path exists and overwrite is False')
        with open(path, 'w') as f:
            f.write(str(self))

    @figure
    def make_plot(
        self,
        axes=None,
        show=False,
        figsize=None,
        **kwargs
    ):

        tria = PolyCollection(self.gr3.coords[self.gr3.elements.triangles])
        quad = PolyCollection(self.gr3.coords[self.gr3.elements.quads])
        tria.set_array(self.values[self.gr3.elements.tri_idxs])
        quad.set_array(self.values[self.gr3.elements.qua_idxs])
        axes.add_collection(tria)
        axes.add_collection(quad)
        return axes


class Fluxflag(Prop):
    """ Class for writing fluxflag.prop file, which is parameter for
        checking volume and salt conservation.
    """
    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, -1)

    def add_region(self, upstream: LinearRing, downstream: LinearRing):

        def get_next_value_pair():
            next_value = np.max(self.values) + 1
            return next_value + 1, next_value
        upstream_val, downstream_val = get_next_value_pair()

        upstream_node_mask = self.gr3.nodes.gdf.within(Polygon(upstream)).to_numpy()
        upstream_elem_idxs = np.where(np.any(upstream_node_mask[self.gr3.elements.array], axis=1))
        self.values[upstream_elem_idxs] = upstream_val

        downstream_node_mask = self.gr3.nodes.gdf.within(Polygon(downstream)).to_numpy()
        downstream_elem_idxs = np.where(np.any(downstream_node_mask[self.gr3.elements.array], axis=1))
        self.values[downstream_elem_idxs] = downstream_val

    @classmethod
    def from_prop_table(cls, hgrid, prop_table):
        obj = cls.default(hgrid)
        for regions in prop_table_to_list_of_tuples(prop_table):
            obj.add_region(*regions)
        return obj


class Tvdflag(Prop):
    """Class for writing tvd.prop file, which specify horizontal regions 
       where upwind or TVD/TVD^2 is used based on the element property values
       (0: upwind; 1: TVD/TVD^2).
    """

    @classmethod
    def default(cls, hgrid):
        return cls.constant(hgrid, 1)

    @classmethod
    def from_geometry(
            cls,
            gr3: Gr3,
            region: Union[Polygon, MultiPolygon],
            inner_value: float = 0,
            outer_value: float = 1,
    ):
        '''inner_value == 0 implies the use of updwind, TVD is the default
        everywhere'''

        if not isinstance(gr3, Gr3):
            raise TypeError(
                f'Argument gr3 must be an instance of type {Gr3}, not type '
                f'{type(gr3)}.')

        if not isinstance(region, (Polygon, MultiPolygon)):
            raise TypeError(
                f'Argument region must be an instance of types {Polygon} or '
                f'{MultiPolygon}, not type {type(region)}.')

        if isinstance(region, Polygon):
            region = [region]

        obj = cls.constant(gr3, np.nan)
        elements_gdf = gr3.elements.gdf
        gdf_in = elements_gdf.geometry.intersects(region)
        obj.values[gdf_in] = inner_value
        outer_indexes = np.setdiff1d(elements_gdf.index, np.where(gdf_in))
        obj.values[outer_indexes] = outer_value
        return obj


def reg2multipoly(file):
    with open(file) as f:
        f.readline()
        npoly = int(f.readline())
        polygons = []
        for _ in range(npoly):
            line = f.readline().split()
            if len(line) == 0:
                break
            nvrt = int(line[0])
            tflag = int(line[1])
            exterior = []
            for _ in range(nvrt):
                exterior.append(tuple(list(map(float, f.readline().split()))))
            interiors = []
            line = f.readline().split()
            if len(line) != 0:
                while tflag == 1:
                    nvrt = int(line[0])
                    tflag = int(line[1])
                    interior = []
                    for _ in range(nvrt):
                        interior.append(tuple(list(map(float, f.readline().split()))))
                    interiors.append(interior)
            polygons.append(Polygon(exterior, interiors))
    return MultiPolygon(polygons)


def prop_table_to_list_of_tuples(file) -> List[Tuple[Polygon, Polygon]]:
    output = []
    file = pathlib.Path(file)
    with open(file) as f:
        lines = f.read()
    for line in lines.split('\n'):
        if ';' not in line:
            break
        reg_upstream, reg_downstream = line.split(';')
        reg_downstream_file = reg_downstream.split(':')[0]
        reg_upstream_file = reg_upstream.split(':')[0]
        reg_upstream_file = reg_upstream_file.strip()
        reg_downstream_file = reg_downstream_file.strip()
        downstream = reg2multipoly(file.parent / reg_downstream_file).geoms[0].exterior
        upstream = reg2multipoly(file.parent / reg_upstream_file).geoms[0].exterior
        output.append((upstream, downstream))
    return output
